import pytest

from kaiso.attributes import Uuid, Incoming, Outgoing
from kaiso.exceptions import MultipleObjectsFound, NoResultFound
from kaiso.relationships import Relationship
from kaiso.types import Entity


@pytest.fixture
def static_types(manager):
    class Contains(Relationship):
        id = Uuid(unique=True)

    class Box(Entity):
        id = Uuid(unique=True)

        contains = Outgoing(Contains)
        contained_within = Incoming(Contains)

    manager.save(Box)
    manager.save(Contains)

    return {
        'Box': Box,
        'Contains': Contains,
    }


def test_rel_attributes(manager, static_types):
    Box = static_types['Box']
    Contains = static_types['Contains']

    box1 = Box()
    box2 = Box()
    contains = Contains(box1, box2)

    manager.save(box1)
    manager.save(box2)
    manager.save(contains)

    assert [b.id for b in box1.contains] == [box2.id]
    assert box2.contained_within.one().id == box1.id


def test_rel_one_missing(manager, static_types):
    Box = static_types['Box']

    box = Box()

    manager.save(box)

    with pytest.raises(NoResultFound):
        box.contains.one()


def test_rel_one_multiple(manager, static_types):
    Box = static_types['Box']
    Contains = static_types['Contains']

    parent = Box()
    child1 = Box()
    child2 = Box()

    contains1 = Contains(parent, child1)
    contains2 = Contains(parent, child2)

    manager.save(parent)
    manager.save(child1)
    manager.save(child2)
    manager.save(contains1)
    manager.save(contains2)

    with pytest.raises(MultipleObjectsFound):
        parent.contains.one()


def test_empty_rel_attributes(manager, static_types):
    Box = static_types['Box']

    box = Box()

    manager.save(box)
    # TODO: should the attr support len()
    assert len(list(box.contains)) == 0
    assert box.contained_within.first() is None


def test_many_children(manager, static_types):
    Box = static_types['Box']
    Contains = static_types['Contains']

    parent = Box()
    child1 = Box()
    child2 = Box()

    contains1 = Contains(parent, child1)
    contains2 = Contains(parent, child2)

    manager.save(parent)
    manager.save(child1)
    manager.save(child2)
    manager.save(contains1)
    manager.save(contains2)

    assert len(list(parent.contains)) == 2
    assert parent.contains.first().id in [child1.id, child2.id]


def test_reference_relationship_itself(manager, static_types):
    Box = static_types['Box']
    Contains = static_types['Contains']

    parent = Box()
    child = Box()

    contains = Contains(parent, child)

    manager.save(parent)
    manager.save(child)
    manager.save(contains)

    fetched = manager.get(Box, id=str(child.id))
    fetched_rel = next(fetched.contained_within.relationships)

    assert fetched_rel.id == contains.id



""" Provides all the exceptions that may be raised.
"""


class ConnectionError(Exception):
    """ Raised when an error occurs connecting to the Neo4j Database
    """


class UniqueConstraintError(Exception):
    """Raised when attempting to create more than one object with the same
       value on an attribute declared to be unique"""


class MultipleObjectsFound(Exception):
    """ Raised when a caller of a RelationshipManager
    expected a single object, but multiple were returned.
    """


class NoResultFound(Exception):
    """ Raised when a call expected at least one object, but none was found.
    """


class UnknownType(Exception):
    """ Raised when trying to deserialise a class that hasn't been
    registered
    """


class TypeNotPersistedError(Exception):
    """Raised when trying to save an instance of a type that is not
    yet persisted """

    def __init__(self, type_id):
        self.type_id = type_id
        super(TypeNotPersistedError, self).__init__(type_id)

    def __str__(self):
        return "Type `{}` not in db".format(self.type_id)


class DeserialisationError(Exception):
    """ Raised when trying to deserialise a dict with no __type__ key """


class TypeAlreadyRegistered(Exception):
    pass


class TypeAlreadyCollected(Exception):
    pass


class CannotUpdateType(Exception):
    """ Raised when trying to update a type defined in code """


class UnsupportedTypeError(Exception):
    """ Raised when trying to interact with a non-Persistable type """


class NoUniqueAttributeError(Exception):
    """ Raised when trying to uniquely identify an object which has no
    unique attributes """



import random
import string

import py2neo
import pytest

from kaiso.attributes import Integer, String
from kaiso.exceptions import UniqueConstraintError
from kaiso.types import Entity, Relationship


@pytest.fixture
def static_types(manager):
    class UniqueThing(Entity):
        id = Integer(unique=True)
        code = String(unique=True)
        extra = String()

    manager.save(UniqueThing)

    return {
        'UniqueThing': UniqueThing,
    }


class TestReplace(object):
    def test_unique_enforced_on_add(self, manager, static_types):
        """ Currently we can't change unique attributes
            (need to figure out how to retain db integrity during such
            changes)
        """
        UniqueThing = static_types['UniqueThing']

        obj1 = UniqueThing(id=1, code='A', extra='lunch')
        manager.save(obj1)

        # This is interpreted as find object with id:1 and change
        # obj.code to 'B' (not "try to create a new object (1, 'B') )
        obj2 = UniqueThing(id=1, code='B', extra='snacks')
        with pytest.raises(NotImplementedError):
            manager.save(obj2)

    def test_replace_no_conflict(self, manager, static_types):
        UniqueThing = static_types['UniqueThing']

        obj1 = UniqueThing(id=1, code='A', extra='lunch')
        manager.save(obj1)

        # should add as no unique clash
        obj2 = UniqueThing(id=2, code='B', extra='snacks')
        manager.save(obj2)

        rows = manager.query('''START n = node:uniquething("id:*")
                                   RETURN n''')
        rows = list(rows)

        assert len(rows) == 2

        assert rows[0][0].id == 1
        assert rows[1][0].id == 2

    def test_no_change(self, manager, static_types):
        UniqueThing = static_types['UniqueThing']

        obj1 = UniqueThing(id=1, code='A', extra='lunch')
        manager.save(obj1)

        # should be a no-op, end up with only one obj in db
        obj2 = UniqueThing(id=1, code='A', extra='lunch')
        manager.save(obj2)

        rows = manager.query('''START n = node:uniquething("id:*")
                                   RETURN n''')
        rows = list(rows)
        assert len(rows) == 1

    def test_unique_fail(self, manager, static_types):
        UniqueThing = static_types['UniqueThing']

        obj1 = UniqueThing(id=1, code='A')
        manager.save(obj1)

        obj2 = UniqueThing(id=2, code='C')
        manager.save(obj2)

        # no way to add this thing without breaking a unique constraint
        with pytest.raises(UniqueConstraintError):
            obj3 = UniqueThing(id=1, code='C')
            manager.save(obj3)

    def test_change_non_unique_field(self, manager, static_types):
        UniqueThing = static_types['UniqueThing']

        obj1 = UniqueThing(id=1, code='A', extra='chocolate')
        manager.save(obj1)

        obj2 = UniqueThing(id=1, code='A', extra='ice cream')
        manager.save(obj2)
        rows = manager.query('''START n = node:uniquething("id:*")
                                   RETURN n''')
        rows = list(rows)
        assert len(rows) == 1
        assert rows[0][0].extra == 'ice cream'

    def test_remove_prop(self, manager, static_types):
        UniqueThing = static_types['UniqueThing']

        obj1 = UniqueThing(id=1, code='A', extra='chocolate')
        manager.save(obj1)

        obj2 = UniqueThing(id=1, code='A')
        manager.save(obj2)
        rows = manager.query('''START n = node:uniquething("id:*")
                                   RETURN n''')
        rows = list(rows)
        assert len(rows) == 1
        assert rows[0][0].extra is None

    def test_no_existing_index(self, manager):
        name = ''.join(random.choice(string.ascii_letters) for _ in range(20))

        # we have no way of removing indexes from the db, so create a new
        # type that we haven't seen before to test the case where
        # the index does not exist
        RandomThing = type(
            name, (Entity,), {'code': String(unique=True)})

        manager.save(RandomThing)
        indexes = manager._conn.get_indexes(py2neo.neo4j.Node)
        assert name.lower() in indexes

    def test_rel_uniqueness(
            self, manager, static_types):
        UniqueThing = static_types['UniqueThing']

        obj1 = UniqueThing(id=1, code='A', extra='lunch')
        obj2 = UniqueThing(id=2, code='B', extra='snacks')
        manager.save(obj1)
        manager.save(obj2)

        class Follows(Relationship):
            id = Integer(unique=True)

        follow_rel1 = Follows(obj1, obj2, id=1)
        manager.save(follow_rel1)

        follow_rel2 = Follows(obj1, obj2, id=1)
        manager.save(follow_rel2)

        result = manager.query('''
            START n = node:uniquething(id="1")
            MATCH n-[r:FOLLOWS]->()
            RETURN r''')

        result = list(result)
        assert len(result) == 1



"""
Helpers for altering the type hierarchy.

"""

from collections import OrderedDict

from kaiso.types import TypeRegistry


def validate_base_change(manager, type_id, new_bases):
    """
    Make sure it would be ok to change the bases of `type_id` to `new_bases`

    If this would result in an inconsistent type hieararchy,
    raise `ValueError`. Otherwise, return None.


    In `manager.get_type_hierarchy`, `type_id` is guaranteed to appear after
    all its base classes, and before all its subclasses.

    To remain consistent, all new_bases must appear before any subclasses
    of type_id

    We make a new list of (type_id, bases) pairs by
    1) removing the old entry for (type_id, (...))
    2) adding a new entry (type_id, new_bases) once we've seen each
       entry in new_bases

    We also discard attributes since we don't care about them for this.

    Finally, we attempt to create the types in this new hierarchy,
    which will throw `TypeError` if we can't maintain consistency.
    """

    # ensure type_id exists
    registry = manager.type_registry

    # ensure type and all new bases exist (this raises UnknownType otherwise)
    registry.get_class_by_id(type_id)
    [registry.get_class_by_id(base) for base in new_bases]

    def new_type_hierarchy(manager):
        """
        We want to switch out the bases in the entry for `type_id`,
        but all bases may not have been seen yet. If so, we defer
        returning `type_id` (and any subclasses, or any other entries
        that are also "waiting for" parent classes to appear.

        If we had transactions, we could start a transaction, change the
        ISA relationships, call get_type_hierarchy() and then roll back
        instead of this.
        """

        awaited_bases = set(new_bases)
        awaited_bases.add(type_id)

        new_type_inserted = False
        deferred_types = OrderedDict()
        deferred_types[type_id] = tuple(new_bases)

        for test_type_id, test_bases, _ in manager.get_type_hierarchy():
            if set(test_bases).intersection(deferred_types):
                deferred_types[test_type_id] = test_bases
                continue

            if test_type_id != type_id:
                yield (test_type_id, test_bases)

            awaited_bases.discard(test_type_id)

            if not new_type_inserted and not awaited_bases:
                new_type_inserted = True
                while deferred_types:
                    yield deferred_types.popitem(last=False)

        if awaited_bases:
            raise ValueError("One of the bases causes an inheritance cycle")

    temp_type_registry = TypeRegistry()
    for test_type_id, bases in new_type_hierarchy(manager):
        bases = tuple(
            temp_type_registry.get_class_by_id(base) for base in bases
        )
        type_name = str(test_type_id)
        try:
            temp_type_registry.create_type(type_name, bases, {})
        except TypeError as ex:
            # bad mro
            raise ValueError(
                "Invalid mro for {} ({})".format(test_type_id, ex)
            )



import pytest

from kaiso.attributes import String
from kaiso.exceptions import TypeAlreadyRegistered, TypeAlreadyCollected
from kaiso.types import Entity, TypeRegistry, get_declaring_class


def test_register_duplicate():

    type_registry = TypeRegistry()

    with pytest.raises(TypeAlreadyRegistered):
        # this declares a dynamic type
        type_registry.create_type('Duplicate', (Entity,), {})
        type_registry.create_type('Duplicate', (Entity,), {})


def test_collect_duplicate():

    # this declares a static class
    class Duplicate(Entity):
        pass

    with pytest.raises(TypeAlreadyCollected):
        # this also declares a static class
        type("Duplicate", (Entity,), {})


def test_get_declaring_class():

    class X(object):
        foo = String()

    class Y(X):
        foo = String()
        bar = String()

    class Z(Y):
        bar = String()

    # prefer subclass
    assert get_declaring_class(X, "foo") == X
    assert get_declaring_class(X, "bar") is None
    assert get_declaring_class(Z, "foo") == Y
    assert get_declaring_class(Z, "bar") == Z

    # prefer parent
    assert get_declaring_class(X, "foo", prefer_subclass=False) == X
    assert get_declaring_class(X, "bar", prefer_subclass=False) is None
    assert get_declaring_class(Z, "foo", prefer_subclass=False) == X
    assert get_declaring_class(Z, "bar", prefer_subclass=False) == Y



from __future__ import unicode_literals

from logging import getLogger
import uuid

from py2neo import cypher, neo4j

from kaiso.attributes import Outgoing, Incoming, String
from kaiso.connection import get_connection
from kaiso.exceptions import (
    UniqueConstraintError, UnknownType, CannotUpdateType, UnsupportedTypeError,
    TypeNotPersistedError, NoResultFound)
from kaiso.queries import (
    get_create_types_query, get_create_relationship_query, get_start_clause,
    join_lines)
from kaiso.references import set_store_for_object
from kaiso.relationships import InstanceOf, IsA, DeclaredOn
from kaiso.serialize import (
    dict_to_db_values_dict, get_changes, object_to_db_value)
from kaiso.types import (
    INTERNAL_CLASS_ATTRS, Descriptor, Persistable, PersistableType,
    Relationship, TypeRegistry, AttributedBase, get_declaring_class,
    get_index_name, get_type_id, is_indexable)
from kaiso.utils import dict_difference


# Note: some Cypher queries contain dummy names for unused nodes, e.g. (dummy1)
# instead of (). This is an attempted workaround for an intermittent Cypher
# bug (https://github.com/neo4j/neo4j/issues/1040) and may be removed when
# used against later versions of Neo4j.


log = getLogger(__name__)


class TypeSystem(AttributedBase):
    """ ``TypeSystem`` is a node that represents the root
    of the type hierarchy.

    Inside the database, the current version of the hierarchy is
    tracked using a ``version`` attribute on the TypeSystem node.
    """
    id = String(unique=True)


def get_attr_filter(obj, type_registry):
    """ Generates a dictionary, that will identify the given ``obj``
    based on it's unique attributes.

    Args:
        obj: An object to look up by index

    Returns:
        A dictionary of key-value pairs to indentify an object by.
    """
    indexes = type_registry.get_index_entries(obj)
    index_filter = dict((key, value) for _, key, value in indexes)
    return index_filter


class Manager(object):
    """Manage the interface to graph-based queryable object store.

    The any object can be saved as long as its type is registered.
    This includes instances of Entity, PersistableType
    and subclasses of either.

    InstanceOf and IsA relationships are automatically generated
    when persisting an object.
    """
    _type_registry_cache = None

    def __init__(self, connection_uri, skip_type_loading=False):
        """ Initializes a Manager object.

        Args:
            connection_uri: A URI used to connect to the graph database.
        """
        self._conn = get_connection(connection_uri)
        self.type_system = TypeSystem(id='TypeSystem')
        self.type_registry = TypeRegistry()
        idx_name = get_index_name(TypeSystem)
        self._conn.get_or_create_index(neo4j.Node, idx_name)
        self.save(self.type_system)
        if not skip_type_loading:
            self.reload_types()

    def _execute(self, query, **params):
        """ Runs a cypher query returning only raw rows of data.

        Args:
            query: A parameterized cypher query.
            params: The parameters used by the query.

        Returns:
            A generator with the raw rows returned by the connection.
        """
        # 2.0 compatibility as we transition
        query = "CYPHER 1.9 {}".format(query)

        log.debug('running query:\n%s', query.format(**params))

        rows, _ = cypher.execute(self._conn, query, params)

        return (row for row in rows)

    def _convert_value(self, value):
        """ Converts a py2neo primitive(Node, Relationship, basic object)
        to an equvalent python object.
        Any value which cannot be converted, will be returned as is.

        Args:
            value: The value to convert.

        Returns:
            The converted value.
        """

        if isinstance(value, (neo4j.Node, neo4j.Relationship)):
            properties = value._properties

            obj = self.type_registry.dict_to_object(properties)

            if isinstance(value, neo4j.Relationship):
                # prefetching start and end-nodes as they don't have
                # their properties loaded yet
                value.start_node.get_properties()
                value.end_node.get_properties()

                obj.start = self._convert_value(value.start_node)
                obj.end = self._convert_value(value.end_node)
            else:
                set_store_for_object(obj, self)
            return obj
        return value

    def _convert_row(self, row):
        for value in row:
            if isinstance(value, list):
                yield [self._convert_value(v) for v in value]
            else:
                yield self._convert_value(value)

    def _index_object(self, obj, node_or_rel):

        indexes = self.type_registry.get_index_entries(obj)
        for index_name, key, value in indexes:
            if isinstance(obj, Relationship):
                index_type = neo4j.Relationship
            else:
                index_type = neo4j.Node

            log.debug(
                'indexing %s for %s using index %s',
                obj, node_or_rel, index_name)

            index = self._conn.get_or_create_index(index_type, index_name)
            index.add(key, value, node_or_rel)

        if not isinstance(obj, Relationship):
            set_store_for_object(obj, self)

    def _type_system_version(self):
        query = join_lines(
            'START',
            get_start_clause(self.type_system, 'ts', self.type_registry),
            'RETURN ts.version?'
        )

        rows = self._execute(query)
        (version,) = next(rows)
        return version

    def invalidate_type_system(self):
        query = join_lines(
            'START',
            get_start_clause(self.type_system, 'ts', self.type_registry),
            'SET ts.version = {new_version}'
        )

        new_version = uuid.uuid4().hex
        next(self._execute(query, new_version=new_version), None)

    def reload_types(self):
        """Reload the type registry for this instance from the graph
        database.
        """
        current_version = self._type_system_version()
        if Manager._type_registry_cache:
            cached_registry, version = Manager._type_registry_cache
            if current_version == version:
                log.debug(
                    'using cached type registry, version: %s', current_version)
                self.type_registry = cached_registry.clone()
                return

        self.type_registry = TypeRegistry()
        registry = self.type_registry

        for type_id, bases, attrs in self.get_type_hierarchy():
            try:
                cls = registry.get_class_by_id(type_id)

                # static types also get loaded into dynamic registry
                # to allow them to be augmented
                if registry.is_static_type(cls):
                    cls = None
            except UnknownType:
                cls = None

            if cls is None:
                bases = tuple(registry.get_class_by_id(base) for base in bases)
                registry.create_type(str(type_id), bases, attrs)

            registry._types_in_db.add(type_id)

        Manager._type_registry_cache = (
            self.type_registry.clone(),
            current_version
        )

    def _get_changes(self, persistable):
        changes = {}
        existing = None
        obj_type = type(persistable)

        registry = self.type_registry

        if isinstance(persistable, PersistableType):
            # this is a class, we need to get it and it's attrs
            idx_name = get_index_name(PersistableType)
            self._conn.get_or_create_index(neo4j.Node, idx_name)

            type_id = get_type_id(persistable)
            query_args = {
                'type_id': type_id
            }

            query = join_lines(
                'START cls=node:%s(id={type_id})' % idx_name,
                'MATCH attr -[:DECLAREDON*0..]-> cls',
                'RETURN cls, collect(attr.name?)'
            )

            # don't use self.query since we don't want to convert the py2neo
            # node into an object
            rows = self._execute(query, **query_args)
            cls_node, attrs = next(rows, (None, None))

            if cls_node is None:
                # have not found the cls
                return None, {}

            existing_cls_attrs = cls_node._properties

            # Make sure we get a clean view of current data.
            registry.refresh_type(persistable)

            new_cls_attrs = registry.object_to_dict(persistable)

            # If any existing keys in "new" are missing in "old", add `None`s.
            # Unlike instance attributes, we just need to remove the properties
            # from the node, which we can achieve by setting the values to None
            for key in set(existing_cls_attrs) - set(new_cls_attrs):
                new_cls_attrs[key] = None
            changes = get_changes(old=existing_cls_attrs, new=new_cls_attrs)

            attrs = set(attrs)

            modified_attrs = {}

            descr = registry.get_descriptor(persistable)
            for name, attr in descr.declared_attributes.items():
                if name not in attrs:
                    modified_attrs[name] = attr

            del_attrs = set(attrs)

            for name in Descriptor(persistable).attributes.keys():
                del_attrs.discard(name)

            for name in del_attrs:
                modified_attrs[name] = None

            if modified_attrs:
                changes['attributes'] = modified_attrs

            # we want to return the existing class
            existing = registry.get_descriptor_by_id(type_id).cls
        else:
            existing = self.get(obj_type, **get_attr_filter(persistable,
                                                            registry))
            if existing is not None:
                existing_props = registry.object_to_dict(existing)
                props = registry.object_to_dict(persistable)

                if existing_props == props:
                    return existing, {}

                changes = get_changes(old=existing_props, new=props)

        return existing, changes

    def _update_types(self, cls):
        query, objects, query_args = get_create_types_query(
            cls, self.type_system, self.type_registry)

        nodes_or_rels = next(self._execute(query, **query_args))

        for obj in objects:
            type_id = get_type_id(obj)
            self.type_registry._types_in_db.add(type_id)
            indexes = self.type_registry.get_indexes_for_type(obj)
            for index_name, _, _ in indexes:
                self._conn.get_or_create_index(neo4j.Node, index_name)

        for obj, node_or_rel in zip(objects, nodes_or_rels):
            self._index_object(obj, node_or_rel)

        # we can't tell whether the CREATE UNIQUE from get_create_types_query
        # will have any effect, so we must invalidate.
        self.invalidate_type_system()
        return cls

    def _update(self, persistable, existing, changes):

        registry = self.type_registry

        for _, index_attr, _ in registry.get_index_entries(existing):
            if index_attr in changes:
                raise NotImplementedError(
                    "We currently don't support changing unique attributes")

        set_clauses = ', '.join([
            'n.%s={%s}' % (key, key) for key, value in changes.items()
            if not isinstance(value, dict)
        ])

        if set_clauses:
            set_clauses = 'SET %s' % set_clauses
        else:
            set_clauses = ''

        if isinstance(persistable, type):

            query_args = {'type_id': get_type_id(persistable)}
            class_attr_changes = {k: v for k, v in changes.items()
                                  if k != 'attributes'}
            query_args.update(class_attr_changes)

            where = []

            descr = registry.get_descriptor(persistable)
            for attr_name in descr.declared_attributes.keys():
                where.append('attr.name = {attr_%s}' % attr_name)
                query_args['attr_%s' % attr_name] = attr_name

            if where:
                where = ' OR '.join(where)
                where = 'WHERE not(%s)' % where
            else:
                where = ''

            index_name = get_index_name(PersistableType)

            query = join_lines(
                'START n=node:%s(id={type_id})' % index_name,
                set_clauses,
                'MATCH attr -[r:DECLAREDON]-> n',
                where,
                'DELETE attr, r',
                'RETURN n',
            )
            self._update_types(persistable)
        else:
            start_clause = get_start_clause(existing, 'n', registry)
            query = None

            if isinstance(persistable, Relationship):
                start_clauses = [start_clause]

                # if start or end have been set, we will do an index lookup
                # to reference them when "updating" the relationship to
                # point to them, if they are not set we look up the original
                # ones using a MATCH clause and "update" the relationship.
                new_start = getattr(persistable, 'start', None)
                new_end = getattr(persistable, 'end', None)

                if new_start is not None:
                    start_clauses.append(
                        get_start_clause(new_start, 'start_node', registry)
                    )
                    match_start_node = '(dummy1)'  # See note at top of page
                else:
                    match_start_node = 'start_node'

                if new_end is not None:
                    start_clauses.append(
                        get_start_clause(new_end, 'end_node', registry)
                    )
                    match_end_node = '(dummy2)'  # See note at top of page
                else:
                    match_end_node = 'end_node'

                start_clause = ', '.join(start_clauses)
                rel_props = registry.object_to_dict(persistable, for_db=True)

                query = join_lines(
                    'START %s' % start_clause,
                    'MATCH %s -[n]-> %s' % (match_start_node, match_end_node),
                    'DELETE n',
                    'CREATE start_node -[r:%s {rel_props}]-> end_node' % (
                        rel_props['__type__'].upper()
                    ),
                    'RETURN r'
                )
                query_args = {'rel_props': rel_props}

            if query is None:
                query = join_lines(
                    'START %s' % start_clause,
                    set_clauses,
                    'RETURN n'
                )
                query_args = changes

        try:
            (result,) = next(self._execute(query, **query_args))
        except StopIteration:
            # this can happen, if no attributes where changed on a type
            result = persistable

        if isinstance(persistable, Relationship):
            self._index_object(persistable, result)
        return result

    def _add(self, obj):
        """ Adds an object to the data store.

        It will automatically generate the type relationships
        for the the object as required and store the object itself.
        """

        query_args = {}
        invalidates_types = False

        if isinstance(obj, PersistableType):
            # object is a type; create the type and its hierarchy
            return self._update_types(obj)

        elif obj is self.type_system:
            query = 'CREATE (n {props}) RETURN n'

        elif isinstance(obj, Relationship):
            # object is a relationship
            obj_type = type(obj)

            if obj_type in (IsA, DeclaredOn):
                invalidates_types = True
            query = get_create_relationship_query(obj, self.type_registry)

        else:
            # object is an instance
            obj_type = type(obj)
            type_id = get_type_id(obj_type)
            if type_id not in self.type_registry._types_in_db:
                raise TypeNotPersistedError(type_id)

            idx_name = get_index_name(PersistableType)
            query = (
                'START cls=node:%s(id={type_id}) '
                'CREATE (n {props}) -[:INSTANCEOF {rel_props}]-> cls '
                'RETURN n'
            ) % idx_name

            query_args = {
                'type_id': get_type_id(obj_type),
                'rel_props': self.type_registry.object_to_dict(
                    InstanceOf(None, None), for_db=True),
            }

        query_args['props'] = self.type_registry.object_to_dict(
            obj, for_db=True)

        (node_or_rel,) = next(self._execute(query, **query_args))
        if invalidates_types:
            self.invalidate_type_system()
        self._index_object(obj, node_or_rel)

        return obj

    def get_type_hierarchy(self, start_type_id=None):
        """ Returns the entire type hierarchy defined in the database
        if start_type_id is None, else returns from that type.

        Returns: A generator yielding tuples of the form
        ``(type_id, bases, attrs)`` where
            - ``type_id`` identifies the type
            - ``bases`` lists the type_ids of the type's bases
            - ``attrs`` lists the attributes defined on the type
        """

        if start_type_id:
            # See note at top of page
            match = ('p=(ts -[:DEFINES]-> (dummy1) <-[:ISA*]- opt '
                     '<-[:ISA*0..]- tpe)')
            where = 'WHERE opt.id = {start_id}'
            query_args = {'start_id': start_type_id}
        else:
            # See note at top of page
            match = 'p=(ts -[:DEFINES]-> (dummy1) <-[:ISA*0..]- tpe)'
            where = ''
            query_args = {}

        query = join_lines(
            'START %s' % get_start_clause(self.type_system, 'ts',
                                          self.type_registry),
            'MATCH',
            match,
            where,
            ''' WITH tpe, max(length(p)) AS level
            MATCH
                tpe <-[?:DECLAREDON*]- attr,
                tpe -[isa?:ISA]-> base

            WITH tpe.id AS type_id, level, tpe AS class_attrs,
                filter(
                    idx_base in collect(DISTINCT [isa.base_index, base.id]):
                        not(LAST(idx_base) is NULL)
                ) AS bases,

                collect(DISTINCT attr) AS attrs

            ORDER BY level
            RETURN type_id, bases, class_attrs, attrs
            ''')

        # we can't use self.query since we don't want to convert the
        # class_attrs dict
        params = dict_to_db_values_dict(query_args)

        for row in self._execute(query, **params):
            type_id, bases, class_attrs, instance_attrs = row

            # the bases are sorted using their index on the IsA relationship
            bases = tuple(base for (_, base) in sorted(bases))
            class_attrs = class_attrs._properties
            for internal_attr in INTERNAL_CLASS_ATTRS:
                class_attrs.pop(internal_attr)
            instance_attrs = [self._convert_value(v) for v in instance_attrs]
            instance_attrs = {attr.name: attr for attr in instance_attrs}

            attrs = class_attrs
            attrs.update(instance_attrs)

            yield (type_id, bases, attrs)

    def serialize(self, obj, for_db=False):
        """ Serialize ``obj`` to a dictionary.

        Args:
            obj: An object to serialize
            for_db: (Optional) bool to indicate whether we are serializing
                data for neo4j or for general transport. This flag propagates
                down all the way into ``Attribute.to_primitive`` and may be
                used by custom attributes to determine behaviour for different
                serialisation targets. E.g. if using a transport that supports
                a Decimal type, `to_primitive` can return Decimal objects if
                for_db is False, and strings otherwise (for persistance in
                the neo4j db).

        Returns:
            A dictionary describing the object
        """
        return self.type_registry.object_to_dict(obj, for_db=for_db)

    def deserialize(self, object_dict):
        """ Deserialize ``object_dict`` to an object.

        Args:
            object_dict: A serialized object dictionary

        Returns:
            An object deserialized using the type registry
        """
        return self.type_registry.dict_to_object(object_dict)

    def create_type(self, name, bases, attrs):
        """ Creates a new class given the name, bases and attrs given.
        """
        return self.type_registry.create_type(name, bases, attrs)

    def update_type(self, tpe, bases):
        """ Change the bases of the given ``tpe``
        """
        if not isinstance(tpe, PersistableType):
            raise UnsupportedTypeError("Object is not a PersistableType")

        if self.type_registry.is_static_type(tpe):
            raise CannotUpdateType("Type '{}' is defined in code and cannot"
                                   "be updated.".format(get_type_id(tpe)))

        descriptor = self.type_registry.get_descriptor(tpe)
        existing_attrs = dict_difference(descriptor.attributes,
                                         descriptor.declared_attributes)
        base_attrs = {}
        for base in bases:
            desc = self.type_registry.get_descriptor(base)
            base_attrs.update(desc.attributes)
        base_attrs = dict_difference(base_attrs,
                                     descriptor.declared_attributes)

        if existing_attrs != base_attrs:
            raise CannotUpdateType("Inherited attributes are not identical")

        start_clauses = [get_start_clause(tpe, 'type', self.type_registry)]
        create_clauses = []
        query_args = {}

        for index, base in enumerate(bases):
            name = 'base_{}'.format(index)
            start = get_start_clause(base, name, self.type_registry)
            create = "type -[:ISA {%s_props}]-> %s" % (name, name)

            query_args["{}_props".format(name)] = {'base_index': index}
            start_clauses.append(start)
            create_clauses.append(create)

        query = join_lines(
            "START",
            (start_clauses, ','),
            "MATCH type -[r:ISA]-> (dummy1)",  # See note at top of page
            "DELETE r",
            "CREATE",
            (create_clauses, ','),
            "RETURN type")

        try:
            next(self._execute(query, **query_args))
            self.invalidate_type_system()
        except StopIteration:
            raise CannotUpdateType("Type or bases not found in the database.")

        self.reload_types()

    def save(self, persistable):
        """ Stores the given ``persistable`` in the graph database.
        If a matching object (by unique keys) already exists, it will
        update it with the modified attributes.
        """
        if not isinstance(persistable, Persistable):
            raise TypeError('cannot persist %s' % persistable)

        existing, changes = self._get_changes(persistable)

        if existing is None:
            self._add(persistable)
            return persistable
        # we always want relationships to go through, even if there
        # are no changes in the properties, e.g. start or end have changed
        elif not changes and not isinstance(persistable, Relationship):
            return persistable
        else:
            return self._update(persistable, existing, changes)

    def save_collected_classes(self, collection):
        classes = collection.values()

        for cls in classes:
            self.type_registry.register(cls)

        for cls in classes:
            self.save(cls)

    def get(self, cls, **attr_filter):

        found = []

        def append_results(query, **query_args):
            rows = self._execute(query, **query_args)
            for (node_or_rel,) in rows:
                found.append(node_or_rel)

        if cls is TypeSystem:
            idx_name = get_index_name(TypeSystem)
            query = join_lines(
                'START ts=node:%s(id={idx_value})' % idx_name,
                'RETURN ts'
            )
            append_results(query, idx_value=self.type_system.id)

        else:
            attr_filter = dict_to_db_values_dict(attr_filter)

            # Workaround that allows nodes and relationships with no
            # attrs to be saved. `save` will cause this method to be called
            # with an empty attr_filter, and when it receives None, will add
            # a new object.
            if not attr_filter:
                return None

            index_found = False
            for index, key, _ in self.type_registry.get_indexes_for_type(cls):
                # ensure this index key is in the filter (to exclude any
                # irrelevant default values set during instantiation)
                if key in attr_filter:
                    index_found = True

                    if issubclass(cls, Relationship):
                        self._conn.get_or_create_index(
                            neo4j.Relationship, index)
                        start_func = 'relationship'
                    else:
                        self._conn.get_or_create_index(neo4j.Node, index)
                        start_func = 'node'

                    idx_val = attr_filter[key]
                    if idx_val is None:
                        raise ValueError(
                            'Cannot `get` using {}=None. None is not an '
                            'indexable value.'.format(key))

                    query = 'START nr = %s:%s(%s={idx_val}) RETURN nr' % (
                        start_func, index, key)
                    append_results(query, idx_val=idx_val)

            if not index_found:
                raise ValueError(
                    'No relevant indexes found when calling get for class: {}'
                    ' with filter {}'.format(cls, attr_filter)
                )

        if not found:
            return None

        # all the nodes returned should be the same
        first = found[0]
        for node in found:
            if node._id != first._id:
                raise UniqueConstraintError((
                    "Multiple nodes ({}) found for unique lookup for "
                    "{}").format(found, cls))

        obj = self._convert_value(first)
        return obj

    def get_by_unique_attr(self, cls, attr_name, values):
        """Bulk load entities from a list of values for a unique attribute

        Returns:
            A generator (obj1, obj2, ...) corresponding to the `values` list

        If any values are missing in the index, the corresponding obj is None
        """

        if not hasattr(cls, attr_name):
            raise ValueError("{} has no attribute {}".format(cls, attr_name))

        attr = getattr(cls, attr_name)
        if not attr.unique:
            raise ValueError("{}.{} is not unique".format(cls, attr_name))

        batch = neo4j.ReadBatch(self._conn)

        declaring_class = get_declaring_class(cls, attr_name)
        index_name = get_index_name(declaring_class)

        for value in values:
            db_value = object_to_db_value(value)
            batch.get_indexed_nodes(index_name, attr_name, db_value)

        # When upgrading to py2neo 1.6, consider changing this to batch.stream
        batch_result = batch.submit()

        def first_or_none(list_):
            return next(iter(list_), None)

        # `batch_result` is a list of either one element lists (for matches)
        # or empty lists. Unpack to flatten (and hydrate to Kaiso objects)
        result = (self._convert_value(
            first_or_none(row)) for row in batch_result)

        return result

    def change_instance_type(self, obj, type_id, updated_values=None):
        if updated_values is None:
            updated_values = {}

        type_registry = self.type_registry

        if type_id not in type_registry._types_in_db:
            raise TypeNotPersistedError(type_id)

        properties = self.serialize(obj, for_db=True)
        properties['__type__'] = type_id
        properties.update(updated_values)

        # get rid of any attributes not supported by the new type
        properties = self.serialize(self.deserialize(properties), for_db=True)

        tpe = type_registry.get_class_by_id(type_id)

        rel_props = type_registry.object_to_dict(InstanceOf(), for_db=True)

        start_clauses = (
            get_start_clause(obj, 'obj', type_registry),
            get_start_clause(tpe, 'tpe', type_registry)
        )

        # See note at top of page
        query = join_lines(
            'START',
            (start_clauses, ','),
            'MATCH (obj)-[old_rel:INSTANCEOF]->(dummy1)',
            'DELETE old_rel',
            'CREATE (obj)-[new_rel:INSTANCEOF {rel_props}]->(tpe)',
            'SET obj={properties}',
            'RETURN obj',
        )

        # use _execute; we need the raw object to add to the index
        results = self._execute(
            query, properties=properties, rel_props=rel_props)
        row = next(results, None)

        if row is None:
            raise NoResultFound(
                "{} not found in db".format(repr(obj))
            )

        (node,) = row
        new_obj = self._convert_value(node)

        # update any indexes
        old_indexes = set(type_registry.get_index_entries(obj))
        new_indexes = set(type_registry.get_index_entries(new_obj))
        indexes_to_remove = old_indexes - new_indexes
        indexes_to_add = new_indexes - old_indexes

        for index_name, key, value in indexes_to_remove:
            index = self._conn.get_index(neo4j.Node, index_name)
            index.remove(key, value)

        for index_name, key, value in indexes_to_add:
            index = self._conn.get_or_create_index(neo4j.Node, index_name)
            index.add(key, value, node)

        set_store_for_object(new_obj, self)

        return new_obj

    def get_related_objects(self, rel_cls, ref_cls, obj):

        if ref_cls is Outgoing:
            rel_query = 'n -[relation:{}]-> related'
        elif ref_cls is Incoming:
            rel_query = 'n <-[relation:{}]- related'

        # TODO: should get the rel name from descriptor?
        rel_query = rel_query.format(rel_cls.__name__.upper())

        query = join_lines(
            'START {idx_lookup} MATCH {rel_query}',
            'RETURN related, relation'
        )

        query = query.format(
            idx_lookup=get_start_clause(obj, 'n', self.type_registry),
            rel_query=rel_query
        )

        return self.query(query)

    def delete(self, obj):
        """ Deletes an object from the store.

        Args:
            obj: The object to delete.

        Returns:
            A tuple: with (number of nodes removed, number of rels removed)
        """
        invalidates_types = False

        if isinstance(obj, Relationship):
            if is_indexable(type(obj)):
                query = join_lines(
                    'START',
                    get_start_clause(obj, 'rel', self.type_registry),
                    'DELETE rel',
                    'RETURN 0, count(rel)'
                )
            else:
                query = join_lines(
                    'START {}, {}',
                    'MATCH n1 -[rel]-> n2',
                    'DELETE rel',
                    'RETURN 0, count(rel)'
                ).format(
                    get_start_clause(obj.start, 'n1', self.type_registry),
                    get_start_clause(obj.end, 'n2', self.type_registry),
                )
            rel_type = type(obj)
            if rel_type in (IsA, DeclaredOn):
                invalidates_types = True

        elif isinstance(obj, PersistableType):
            query = join_lines(
                'START {}',
                'MATCH attr -[?:DECLAREDON]-> obj',
                'DELETE attr',
                'MATCH obj -[rel]- (dummy1)',  # See note at top of page
                'DELETE obj, rel',
                'RETURN count(obj), count(rel)'
            ).format(
                get_start_clause(obj, 'obj', self.type_registry)
            )
            invalidates_types = True
        else:
            query = join_lines(
                'START {}',
                'MATCH obj -[rel]- (dummy1)',  # See note at top of page
                'DELETE obj, rel',
                'RETURN count(obj), count(rel)'
            ).format(
                get_start_clause(obj, 'obj', self.type_registry)
            )

        # TODO: delete node/rel from indexes
        res = next(self._execute(query))
        if invalidates_types:
            self.invalidate_type_system()
        return res

    def query(self, query, **params):
        """ Queries the store given a parameterized cypher query.

        Args:
            query: A parameterized cypher query.
            params: query: A parameterized cypher query.

        Returns:
            A generator with tuples containing stored objects or values.

        WARNING: If you use this method to modify the type hierarchy (i.e.
        types, their declared attributes or their relationships), ensure
        to call ``manager.invalidate_type_hierarchy()`` afterwards.
        Otherwise managers will continue to use cached versions. Instances can
        be modified without changing the type hierarchy.
        """
        params = dict_to_db_values_dict(params)
        result = self._execute(query, **params)

        return (tuple(self._convert_row(row)) for row in result)

    def destroy(self):
        """ Removes all nodes, relationships and indexes in the store. This
            object will no longer be usable after calling this method.
            Construct a new Manager to re-initialise the database for kaiso.

            WARNING: This will destroy everything in your Neo4j database.

        """
        self._conn.clear()
        for index_name in self._conn.get_indexes(neo4j.Node).keys():
            self._conn.delete_index(neo4j.Node, index_name)
        for index_name in self._conn.get_indexes(neo4j.Relationship).keys():
            self._conn.delete_index(neo4j.Relationship, index_name)



from weakref import WeakKeyDictionary


_object_storage_map = WeakKeyDictionary()


def set_store_for_object(obj, store):
    _object_storage_map[obj] = store


def get_store_for_object(obj):
    return _object_storage_map[obj]



from kaiso.utils import dict_difference


def test_dict_difference():

    a = {1: 2, 3: 4}
    b = {3: 4, 5: 6}

    assert dict_difference(a, b) == {1: 2}
    assert dict_difference(b, a) == {5: 6}

    assert dict_difference(a, {}) == a
    assert dict_difference({}, a) == {}

    # non-hashable value
    c = {7: set()}

    assert dict_difference(a, c) == a

    # same keys
    d = {1: 2}
    e = {1: "2"}

    assert dict_difference(d, e) == d



from kaiso.attributes import (
    Bool, Choice, DateTime, Decimal, Float, Integer, String, Tuple, Uuid)


def test_primitive_ignores_none():
    assert Bool.to_primitive(None, True) is None
    assert Choice.to_primitive(None, True) is None
    assert DateTime.to_primitive(None, True) is None
    assert Decimal.to_primitive(None, True) is None
    assert Float.to_primitive(None, True) is None
    assert Integer.to_primitive(None, True) is None
    assert String.to_primitive(None, True) is None
    assert Tuple.to_primitive(None, True) is None
    assert Uuid.to_primitive(None, True) is None
    assert Uuid.to_primitive(None, True) is None


def test_python_ignores_none():
    assert Bool.to_python(None) is None
    assert Choice.to_python(None) is None
    assert DateTime.to_python(None) is None
    assert Decimal.to_python(None) is None
    assert Float.to_python(None) is None
    assert Integer.to_python(None) is None
    assert String.to_python(None) is None
    assert Tuple.to_python(None) is None
    assert Uuid.to_python(None) is None
    assert Uuid.to_python(None) is None



from kaiso import types


class TemporaryStaticTypes(object):
    """ Testing helper to define temporary static types

    Types defined inside this context only live for the duration of a test,
    and so won't have have side effects leaking into other tests

    Usage:
        with TemporaryStaticTypes():
            class Foo(Entity):
                pass

    Though generally via the funcarg

        @pytest.fixture
        def foo(temporary_static_types):
            class Foo(Entity):
                pass
            return Foo

    or
        def test_foo(temporary_static_types):
            class Foo(Entity):
                pass

            assert(...)
    """

    def start(self):
        self.state = types.collected_static_classes.dump_state()

    def stop(self):
        types.collected_static_classes.load_state(self.state)

    def __enter__(self):
        self.start()

    def __exit__(self, *args, **kwargs):
        self.stop()



import pytest

from kaiso.attributes import String
from kaiso.exceptions import UnknownType
from kaiso.types import Entity, TypeRegistry, collector


def test_unknown_type(type_registry):
    with pytest.raises(UnknownType):
        type_registry.get_descriptor(UnknownType)


def test_static_descriptor_caching(manager):
    with collector():

        class Thing(Entity):
            prop_x = String(required=True)

        class That(Entity):
            prop_y = String(unique=True)

        type_registry = TypeRegistry()

        thing_descriptor1 = type_registry.get_descriptor(Thing)
        thing_descriptor2 = type_registry.get_descriptor(Thing)

        that_descriptor1 = type_registry.get_descriptor(That)
        that_descriptor2 = type_registry.get_descriptor(That)

        # assert repeated calls return the same objects
        assert thing_descriptor1 is thing_descriptor2
        assert that_descriptor1 is that_descriptor2

        # check that different types still get different descriptors
        assert thing_descriptor1 is not that_descriptor1


def test_descriptor_property_caching(manager):
    with collector():

        class Thing(Entity):
            prop_x = String(required=True)

        class That(Entity):
            prop_y = String(unique=True)

        type_registry = TypeRegistry()

        thing_descriptor = type_registry.get_descriptor(Thing)
        that_descriptor = type_registry.get_descriptor(That)

        property_names = [
            'attributes', 'relationships', 'declared_attributes',
            'class_attributes', 'declared_class_attributes'
        ]
        for name in property_names:
            thing_val1 = getattr(thing_descriptor, name)
            thing_val2 = getattr(thing_descriptor, name)
            that_val = getattr(that_descriptor, name)
            assert thing_val1 is thing_val2, name
            assert thing_val2 is not that_val, name



def dict_difference(d1, d2):
    """ Returns the difference between two dictionaries.

    Keys and values must be equal.
    """
    sentinel = object()
    return dict(((k, v) for (k, v) in d1.iteritems()
                 if d2.get(k, sentinel) != v))



import pytest

from kaiso.attributes import String
from kaiso.exceptions import UnknownType
from kaiso.types import Entity


def test_static_types_can_be_augmented(manager):
    # make sure static-type Entity is in the graph
    manager.save(Entity)

    # create an entity instance, attempting to set a non-defined attribute,
    # and check that the attribute has not been saved to the graph
    entity1 = Entity(foo="foo")
    manager.save(entity1)
    rows = manager.query("""
        START n=node:persistabletype(id="Entity")
        MATCH n <-[:INSTANCEOF]- entity
        RETURN entity
    """)
    (loaded1, ) = next(rows)
    assert not hasattr(loaded1, 'foo')

    # Define and save a new Entity attribute.
    Entity.foo = String()
    manager.save(Entity)
    # Delete attribute and let `reload_types` restore it
    del Entity.foo
    manager.reload_types()

    # Create an entity with a value for the new attribute.
    # Check that it has now been stored in the graph.
    entity2 = Entity(foo="foo")
    manager.save(entity2)
    rows = manager.query("""
    START n=node:persistabletype(id="Entity")
        MATCH n <-[:INSTANCEOF]- entity
        WHERE entity.foo! = "foo"
        RETURN entity
    """)
    rows = list(rows)
    assert len(rows) == 1
    (loaded2, ) = rows[0]
    assert loaded2.foo == 'foo'


def test_save_dynamic_type(manager):

    attrs = {'id': String(unique=True)}
    Foobar = manager.create_type('Foobar', (Entity,), attrs)

    manager.save(Foobar)

    rows = manager.query('START n=node:persistabletype(id="Foobar") RETURN n')
    (result,) = next(rows)

    assert result is Foobar


def test_save_dynamic_typed_obj(manager):

    attrs = {'id': String(unique=True)}
    Foobar = manager.create_type('Foobar', (Entity,), attrs)
    manager.save(Foobar)

    foo = Foobar(id='spam')
    manager.save(foo)

    rows = manager.query('START n=node:foobar(id="spam") RETURN n')
    (result,) = next(rows)

    assert result.id == foo.id


def test_add_attr_to_type(manager):
    Foobar = manager.create_type('Foobar', (Entity,), {})
    manager.save(Foobar)

    Foobar.ham = String(default='eggs')
    manager.save(Foobar)

    rows = manager.query(
        'START n=node:persistabletype(id="Foobar") '
        'MATCH n <-[:DECLAREDON]- attr '
        'RETURN count(attr)')
    (count,) = next(rows)
    assert count == 1


def test_remove_attr_from_type(manager):
    attrs = {'ham': String()}
    Foobar = manager.create_type('Foobar', (Entity,), attrs)
    manager.save(Foobar)

    del Foobar.ham

    manager.save(Foobar)

    rows = manager.query(
        'START n=node:persistabletype(id="Foobar") '
        'MATCH n <-[:DECLAREDON]- attr '
        'RETURN count(attr)')
    (count,) = next(rows)
    assert count == 0


def test_removing_attr_from_declared_type_does_not_remove_it(manager):

    # the use case:
    # developer defines a partial type in code
    # the type Ham gets updated in the DB at runtime
    # then the code changes, removing an attribute from the class
    # the type definition in the DB is not affected, nor
    # the data gets back from the DB

    class Ham(Entity):
        egg = String

    manager.save(Ham)

    attrs = {'egg': String(), 'spam': String()}
    DynHam = manager.create_type('Ham', (Entity,), attrs)
    manager.save(DynHam)

    del Ham.egg
    manager.save(Ham)

    rows = manager.query(
        'START n=node:persistabletype(id="Ham") '
        'MATCH n <-[:DECLAREDON]- attr '
        'RETURN count(attr)')
    (count,) = next(rows)
    assert count == 2


def test_load_dynamic_types(manager):
    Animal = manager.create_type('Animal', (Entity,), {'id': String()})
    Horse = manager.create_type('Horse', (Animal,), {'hoof': String()})
    Duck = manager.create_type('Duck', (Animal,), {'beek': String()})
    Beaver = manager.create_type('Beaver', (Animal,), {'tail': String()})
    Platypus = manager.create_type(
        'Platypus', (Duck, Beaver), {'egg': String()})

    manager.save(Horse)
    manager.save(Platypus)

    # this is the same as creating a new manager
    manager.reload_types()

    rows = manager.query(
        '''
        START ts=node:typesystem(id="TypeSystem")
        MATCH p = (ts -[:DEFINES]-> () <-[:ISA*0..]- tpe),
            tpe <-[:DECLAREDON*0..]- attr,
            tpe -[:ISA*0..1]-> base
        RETURN tpe.id,  length(p) AS level,
            filter(b_id in collect(distinct base.id): b_id <> tpe.id),
            collect(distinct attr.name?)
        ORDER BY level, tpe.id
        ''')
    result = list(rows)

    assert result == [
        ('Entity', 1, [], []),
        ('Animal', 2, ['Entity'], ['id']),
        ('Beaver', 3, ['Animal'], ['tail']),
        ('Duck', 3, ['Animal'], ['beek']),
        ('Horse', 3, ['Animal'], ['hoof']),
        ('Platypus', 4, ['Duck', 'Beaver'], ['egg']),
    ]


def test_add_attr_to_type_via_2nd_manager(manager):
    attrs = {'id': String(unique=True)}
    Shrub = manager.create_type('Shrub', (Entity,), attrs)
    manager.save(Shrub)

    shrub = Shrub(id='spam')
    manager.save(shrub)

    # this is the same as creating a new manager using the same URL
    manager.reload_types()

    (Shrub,) = next(manager.query(
        'START cls=node:persistabletype(id="Shrub") RETURN cls'))
    Shrub.newattr = String(default='eggs')
    manager.save(Shrub)

    # we want to query from an independent manager
    manager.reload_types()
    rows = manager.query('START n=node:shrub(id="spam") RETURN n')
    (result,) = next(rows)

    assert result.newattr is None


def test_type_registry_independence(manager):
    Shrub = manager.create_type('Shrub', (Entity,), {})
    manager.type_registry.get_descriptor(Shrub)

    manager.reload_types()
    with pytest.raises(UnknownType):
        manager.type_registry.get_descriptor(Shrub)



""" Provides Relationship classes to be used by client code.

All relationship classes must inherit from Relationship
to be usable within the framework.

The IsA and InstanceOf relationships are used internally for describing
inheritance relationships and are autogenerated when an object is added
to the data store.

They may be used for querying and map to pyhton's issubclass()
and isinstanceof() functionality.
"""
from kaiso.types import Relationship
from kaiso.attributes import Integer


class IsA(Relationship):
    """ Describes a sub-class relationship between two classes.

    It maps to python's issubclass(sub_class, base_class) as
    IsA(sub_class, base_class).
    """
    base_index = Integer(required=True)


class InstanceOf(Relationship):
    """ Describes a instance to type relationship
    between an object and a class.

    It maps to python's isinstance(obj, the_class) as
    InstanceOf(obj, the_class).
    """


class DeclaredOn(Relationship):
    """ Describes the relationship between a type and the properties that
    instances of that type can have.
    """


class Defines(Relationship):
    """ Relationship used by the ``TypeSystem`` to connect to the types
    it defines.
    """



import decimal
import uuid
import datetime
import iso8601

from kaiso.attributes.bases import RelationshipReference, wraps_type
from kaiso.types import Attribute, DefaultableAttribute


class Outgoing(RelationshipReference):
    pass


class Incoming(RelationshipReference):
    pass


@wraps_type(uuid.UUID)
class Uuid(Attribute):
    @property
    def default(self):
        return uuid.uuid4()

    @staticmethod
    def to_primitive(value, for_db):
        if value is None:
            return None
        return str(value)

    @staticmethod
    def to_python(value):
        if value is None:
            return None
        return uuid.UUID(hex=value)


class Bool(DefaultableAttribute):
    pass


class Integer(DefaultableAttribute):
    pass


class Float(DefaultableAttribute):
    pass


class String(DefaultableAttribute):
    pass


class Tuple(DefaultableAttribute):
    @staticmethod
    def to_primitive(value, for_db):
        if value is None:
            return None
        return list(value)

    @staticmethod
    def to_python(value):
        if value is None:
            return None
        return tuple(value)


@wraps_type(decimal.Decimal)
class Decimal(DefaultableAttribute):
    @staticmethod
    def to_primitive(value, for_db):
        if value is None:
            return None
        return str(value)

    @staticmethod
    def to_python(value):
        if value is None:
            return None
        return decimal.Decimal(value)


@wraps_type(datetime.datetime)
class DateTime(DefaultableAttribute):
    @staticmethod
    def to_primitive(value, for_db):
        if value is None:
            return None
        return value.isoformat()

    @staticmethod
    def to_python(value):
        if value is None:
            return None
        return iso8601.parse_date(value)


class Choice(DefaultableAttribute):
    choices = Tuple(default=tuple(), required=True)

    def __init__(self, *choices, **kwargs):
        super(Choice, self).__init__(**kwargs)
        self.choices = choices



import os
import socket
import subprocess
import shutil

from mock import patch, ANY
import pytest

import kaiso.connection
from kaiso.connection import get_connection, write_config
from kaiso.exceptions import ConnectionError


@pytest.mark.slow
class TestTempConnectionProcesses():
    """ Test spinning up temporary neo4j processes.

    We mock out ``py2neo.neo4j.GraphDatabaseService`` because:
        a) it's a library and we're not testing its behaviour directly
        b) the first thing GraphDatabaseService does is generate a
           GET request (the same one we do in ``get_connection``)
           whch sometimes results in a SocketError. This is possibly
           a bug in neo4j, but it shouldn't influence these tests.
    """
    temp_dir = '/tmp/foo'

    def setup_method(self, method):
        # patch write_config to itself so we can assert arguments
        self._p_write_config = patch('kaiso.connection.write_config')
        self.write_config = self._p_write_config.start()
        self.write_config.side_effect = write_config

        self._original_temporary_databases = (
            kaiso.connection._temporary_databases.copy())
        kaiso.connection._temporary_databases = {}

    def teardown_method(self, method):
        # kill processes
        for key, temp_neo4j in kaiso.connection._temporary_databases.items():
            temp_neo4j.process.terminate()
            temp_neo4j.process.wait()
            del kaiso.connection._temporary_databases[key]
        # remove temp directory
        if os.path.exists(self.temp_dir):
            shutil.rmtree(self.temp_dir)
        # stop patch
        self._p_write_config.stop()

        kaiso.connection._temporary_databases = (
            self._original_temporary_databases)

    def test_temp_connection_defaults(self):
        """ Verify temporary connection with default options.
        """
        with patch('py2neo.neo4j.GraphDatabaseService', lambda uri: uri):
            # make sure we don't clash with the default temp port
            with patch('kaiso.connection.DEFAULT_TEMP_PORT', 7776):
                conn = get_connection('temp://')

        assert conn == "http://localhost:7776/db/data/"
        self.write_config.assert_called_once_with(
            ANY, 7776, ANY, 'localhost'
        )

    def test_temp_connection_custom_port(self):
        """ Verify temporary connection on a custom port.
        """
        port = "7777"
        with patch('py2neo.neo4j.GraphDatabaseService', lambda uri: uri):
            conn = get_connection('temp://:{}'.format(port))

        assert conn == "http://localhost:{}/db/data/".format(port)
        self.write_config.assert_called_once_with(
            ANY, 7777, ANY, 'localhost'
        )

    def test_temp_connection_custom_data_dir(self):
        """ Verify temporary connection with a custom data directory.
        """
        temp_dir = self.temp_dir
        with patch('py2neo.neo4j.GraphDatabaseService', lambda uri: uri):
            conn = get_connection('temp://{}'.format(temp_dir))

        assert conn == "http://localhost:7475/db/data/"
        assert os.path.exists(temp_dir)
        self.write_config.assert_called_once_with(
            os.path.join(temp_dir, 'server.properties'), 7475,
            os.path.join(temp_dir, 'data'), 'localhost'
        )

    def test_temp_connection_custom_bind_iface(self):
        """ Verify temporary connection with bound to a specific interface.
        """
        bind_iface = "0.0.0.0"
        with patch('py2neo.neo4j.GraphDatabaseService', lambda uri: uri):
            conn = get_connection('temp://{}'.format(bind_iface))

        assert conn == "http://localhost:7475/db/data/"
        self.write_config.assert_called_once_with(
            ANY, 7475, ANY, bind_iface
        )

    def test_temp_connection_custom(self):
        """ Verify temporary connection with custom everything.
        """
        port = "7777"
        bind_iface = socket.gethostname().lower()
        temp_dir = self.temp_dir
        with patch('py2neo.neo4j.GraphDatabaseService', lambda uri: uri):
            conn = get_connection('temp://{}:{}{}'.format(bind_iface, port,
                                                          temp_dir))

        assert conn == "http://{}:{}/db/data/".format(bind_iface, port)
        assert os.path.exists(temp_dir)
        self.write_config.assert_called_once_with(
            os.path.join(temp_dir, 'server.properties'), 7777,
            os.path.join(temp_dir, 'data'), bind_iface
        )

    def test_multiple_temp_connections(self):
        """ Verify that temporary connections are recycled if the uri matches
        """
        with patch('py2neo.neo4j.GraphDatabaseService', lambda uri: uri):
            conn1 = get_connection('temp://')
            with patch.object(kaiso.connection, 'get_neo4j_info') as get_info:
                conn2 = get_connection('temp://')
                assert conn1 == conn2
                assert not get_info.called  # we should be reusing existing db

    def test_temp_connection_timeout(self):
        """ Verify that an exception is raised if the process times out.
        """
        with patch.object(kaiso.connection, 'TIMEOUT', 0):
            with pytest.raises(ConnectionError):
                get_connection('temp://8888')


def test_temp_connection_no_neo4j_info():
    """ Verify that we throw an error if neo4j info cannot be determined.
    """
    with patch.object(subprocess, 'check_output') as check_output:
        check_output.side_effect = OSError()
        with pytest.raises(ConnectionError):
            get_connection('temp://8888')


def test_temp_connection_empty_string():
    """ Verify that we throw an error if an empty connection string is used.
    """
    with pytest.raises(ConnectionError):
        get_connection('')



from mock import ANY
import pytest

from kaiso.types import Entity
from kaiso.attributes import Uuid, String, Bool


@pytest.fixture
def static_types(manager):

    class FooType(Entity):
        id = Uuid(unique=True)
        cls_attr = "placeholder"

    class BarType(Entity):
        extra = String()
        cls_attr = "placeholder"

    class BazType(Entity):
        name = String()
        special = Bool()
        cls_attr = "placeholder"

    class AType(Entity):
        foo = String()

    class BType(AType):
        pass

    return {
        'FooType': FooType,
        'BarType': BarType,
        'BazType': BazType,
        'AType': AType,
        'BType': BType,
    }


def test_get_clsss_by_id_returns_static_type(type_registry, static_types):
    FooType = static_types['FooType']

    # create a dynamic FooType
    attrs = {'id': Uuid(unique=True), 'extra': String(unique=True)}
    type_registry.create_type("FooType", (), attrs)

    # check that cls_id 'FooType' gives us the static FooType
    assert type_registry.get_class_by_id("FooType") == FooType


def test_get_descriptor_returns_dynamic_type(type_registry, static_types):

    # create a dynamic FooType
    attrs = {'id': Uuid(unique=True), 'extra': String(unique=True)}
    DynamicFooType = type_registry.create_type("FooType", (), attrs)

    # check that the descriptor for FooType is the same as the descriptor for
    # DynamicFooType
    dynamic_descriptor = type_registry.get_descriptor(DynamicFooType)
    assert type_registry.get_descriptor_by_id("FooType") == dynamic_descriptor

    # check that the descriptor contains DynamicFooType's extra attributes
    assert "extra" in type_registry.get_descriptor_by_id("FooType").attributes


def test_get_index_entries(type_registry, static_types):
    FooType = static_types['FooType']

    # create a dynamic FooType
    attrs = {'id': Uuid(unique=True), 'extra': String(unique=True)}
    DynamicFooType = type_registry.create_type("FooType", (), attrs)

    # check index entries for the types
    assert list(type_registry.get_index_entries(FooType)) == [
        ('persistabletype', 'id', 'FooType')
    ]
    assert list(type_registry.get_index_entries(DynamicFooType)) == [
        ('persistabletype', 'id', 'FooType')
    ]

    # create an instance
    foo = DynamicFooType()
    foo.extra = "hello"

    # check that indices are returned for both unique attributes
    # checl that 'extra' has a value
    index_entries = list(type_registry.get_index_entries(foo))
    assert len(index_entries) == 2
    assert index_entries[0][0] == "footype"
    assert index_entries[0][1] == "id"
    assert index_entries[0][2] == ANY  # id not known
    assert index_entries[1][0] == "footype"
    assert index_entries[1][1] == "extra"
    assert index_entries[1][2] == "hello"


def test_is_static_type(type_registry, static_types):

    # test purely static type
    FooType = static_types['FooType']
    assert type_registry.is_static_type(FooType) is True

    # test purely dynamic type
    NewType = type_registry.create_type("NewType", (Entity,), {})
    assert type_registry.is_static_type(NewType) is False

    # augment a static type
    attrs = {'id': Uuid(unique=True), 'extra': String(unique=True)}
    AugmentedFooType = type_registry.create_type("FooType", (), attrs)

    assert type_registry.is_static_type(AugmentedFooType) is False
    assert type_registry.is_static_type(FooType) is True


def test_has_code_defined_attribute(type_registry, static_types):
    attrs = {
        'id': Uuid(unique=True),
        'extra': String(unique=True),
        'cls_attr': "placeholder"
    }

    # test purely dynamic type
    NewType = type_registry.create_type("NewType", (Entity,), attrs)
    assert not type_registry.has_code_defined_attribute(NewType, "extra")
    assert not type_registry.has_code_defined_attribute(NewType, "cls_attr")
    assert not type_registry.has_code_defined_attribute(NewType, "nonexistent")

    # test purely code-defined type
    BarType = static_types['BarType']

    assert type_registry.has_code_defined_attribute(BarType, "extra")
    assert type_registry.has_code_defined_attribute(BarType, "cls_attr")
    assert not type_registry.has_code_defined_attribute(BarType, "nonexistent")

    # augment FooType with an "extra" attr; redefine "id" and "cls_attr" attrs
    FooType = static_types['FooType']
    type_registry.create_type("FooType", (Entity,), attrs)

    # test augmented type
    assert type_registry.has_code_defined_attribute(FooType, "id")
    assert not type_registry.has_code_defined_attribute(FooType, "extra")
    assert type_registry.has_code_defined_attribute(FooType, "cls_attr")
    assert not type_registry.has_code_defined_attribute(FooType, "nonexistent")

    # create static type with a dynamic subclass
    BazType = static_types['BazType']

    attrs = {'special': Bool(), 'extra': Bool(), 'cls_attr': "placeholder"}
    SubBazType = type_registry.create_type("SubBazType", (BazType,), attrs)

    # test the static type
    assert type_registry.has_code_defined_attribute(BazType, "name")
    assert type_registry.has_code_defined_attribute(BazType, "special")
    assert type_registry.has_code_defined_attribute(BazType, "cls_attr")

    # test the subtype does not override it
    assert type_registry.has_code_defined_attribute(SubBazType, "name")
    assert type_registry.has_code_defined_attribute(SubBazType, "special")
    assert type_registry.has_code_defined_attribute(SubBazType, "cls_attr")
    assert not type_registry.has_code_defined_attribute(SubBazType, "extra")

    # test augmented type after reload
    AType = static_types['AType']
    BType = static_types['BType']

    # augment the types
    # A1 and B1 are the new types A and B would deserialize to
    # after reloading the type hierarchy
    A1Type = type_registry.create_type("AType", (Entity,), {'bar': True})
    B1Type = type_registry.create_type("BType", (Entity,), {'bar': True})

    assert type_registry.has_code_defined_attribute(AType, "foo")
    assert type_registry.has_code_defined_attribute(BType, "foo")
    assert type_registry.has_code_defined_attribute(A1Type, "foo")
    assert type_registry.has_code_defined_attribute(B1Type, "foo")

    assert not type_registry.has_code_defined_attribute(AType, "bar")
    assert not type_registry.has_code_defined_attribute(BType, "bar")
    assert not type_registry.has_code_defined_attribute(A1Type, "bar")
    assert not type_registry.has_code_defined_attribute(B1Type, "bar")



from functools import wraps


def unique(fn):
    """ Wraps a function to return only unique items.
    The wrapped function must return an iterable object.
    When the wrapped function is called, each item from the iterable
    will be yielded only once and duplicates will be ignored.

    Args:
        fn: The function to be wrapped.

    Returns:
        A wrapper function for fn.
    """
    @wraps(fn)
    def wrapped(*args, **kwargs):
        items = set()
        for item in fn(*args, **kwargs):
            if item not in items:
                items.add(item)
                yield item
    return wrapped



from kaiso.exceptions import NoUniqueAttributeError
from kaiso.relationships import InstanceOf, IsA, DeclaredOn, Defines
from kaiso.types import (
    AttributedBase, get_index_name, Relationship)
from kaiso.serialize import get_type_relationships


def join_lines(*lines, **kwargs):
    rows = []
    sep = kwargs.get('sep', '\n')

    for lne in lines:
        if isinstance(lne, tuple):
            (lne, s) = lne
            lne = '    ' + join_lines(sep=s + '\n    ', *lne)

        if lne != '':
            rows.append(lne)

    return sep.join(rows)


def get_start_clause(obj, name, type_registry):
    """ Returns a node lookup by index as used by the START clause.

    Args:
        obj: An object to create an index lookup.
        name: The name of the object in the query.
    Returns:
        A string with index lookup of a cypher START clause.
    """

    index = next(type_registry.get_index_entries(obj), None)
    if index is None:
        raise NoUniqueAttributeError()

    if isinstance(obj, Relationship):
        index_type = "rel"
    else:
        index_type = "node"
    query = '{}={}:{}({}="{}")'.format(name, index_type, *index)
    return query


def get_create_types_query(cls, root, type_registry):
    """ Returns a CREATE UNIQUE query for an entire type hierarchy.

    Includes statements that create each type's attributes.

    Args:
        cls: An object to create a type hierarchy for.

    Returns:
        A tuple containing:
        (cypher query, classes to create nodes for, the object names).
    """
    hierarchy_lines = []
    set_lines = []
    classes = {}

    query_args = {
        'root_id': root.id,
        'Defines_props': type_registry.object_to_dict(Defines()),
        'InstanceOf_props': type_registry.object_to_dict(InstanceOf()),
        'DeclaredOn_props': type_registry.object_to_dict(DeclaredOn()),
    }

    # filter type relationships that we want to persist
    type_relationships = []
    for cls1, rel_cls_idx, cls2 in get_type_relationships(cls):
        if issubclass(cls2, AttributedBase):
            type_relationships.append((cls1, rel_cls_idx, cls2))

    # process type relationships
    is_first = True
    isa_props_counter = 0

    for cls1, (rel_cls, base_idx), cls2 in type_relationships:

        name1 = cls1.__name__

        if name1 in classes:
            abstr1 = '`%s`' % (name1,)
        else:
            abstr1 = '(`%s` {%s_id_props})' % (name1, name1)

        classes[name1] = cls1

        if is_first:
            is_first = False
            ln = 'root -[:DEFINES {Defines_props}]-> %s' % abstr1
        else:
            name2 = cls2.__name__
            classes[name2] = cls2

            rel_name = rel_cls.__name__
            rel_type = rel_name.upper()

            prop_name = "%s_props" % rel_name

            if rel_cls is IsA:
                prop_name = '%s_%d' % (rel_name, isa_props_counter)
                isa_props_counter += 1

                props = type_registry.object_to_dict(IsA(base_index=base_idx))
                query_args[prop_name] = props

            ln = '%s -[%s:%s]-> `%s`' % (abstr1, prop_name, rel_type, name2)
            set_lines.append('SET `%s` = {%s}' % (prop_name, prop_name))
        hierarchy_lines.append(ln)

    # process attributes
    for name, cls in classes.items():

        descriptor = type_registry.get_descriptor(cls)
        attributes = descriptor.declared_attributes
        for attr_name, attr in attributes.iteritems():
            key = "%s_%s" % (name, attr_name)

            ln = '({%s}) -[:DECLAREDON {DeclaredOn_props}]-> `%s`' % (
                key, name)
            hierarchy_lines.append(ln)

            attr_dict = type_registry.object_to_dict(
                attr, for_db=True)

            attr_dict['name'] = attr_name
            query_args[key] = attr_dict

    # processing class attributes
    for key, cls in classes.iteritems():
        # all attributes of the class to be set via the query
        cls_props = type_registry.object_to_dict(cls, for_db=True)
        query_args['%s_props' % key] = cls_props
        set_lines.append('SET `%s` = {%s_props}' % (key, key))

        # attributes which uniquely identify the class itself
        # these are used in the CREATE UNIQUE part of the query
        cls_id_props = {
            '__type__': cls_props['__type__'],
            'id': cls_props['id']
        }
        query_args['%s_id_props' % key] = cls_id_props

    quoted_names = ('`{}`'.format(cls) for cls in classes.keys())
    query = join_lines(
        'START root=node:%s(id={root_id})' % get_index_name(type(root)),
        'CREATE UNIQUE',
        (hierarchy_lines, ','),
        (set_lines, ''),
        'RETURN %s' % ', '.join(quoted_names)
    )

    return query, classes.values(), query_args


def get_create_relationship_query(rel, type_registry):
    rel_props = type_registry.object_to_dict(rel, for_db=True)
    query = 'START %s, %s CREATE n1 -[r:%s {props}]-> n2 RETURN r'

    query = query % (
        get_start_clause(rel.start, 'n1', type_registry),
        get_start_clause(rel.end, 'n2', type_registry),
        rel_props['__type__'].upper(),
    )

    return query



from __future__ import absolute_import  # local types.py and builtin types

from contextlib import contextmanager
from functools import wraps
from inspect import getmembers, getmro

from kaiso.exceptions import (
    UnknownType, TypeAlreadyRegistered, TypeAlreadyCollected,
    DeserialisationError)


# at some point, rename id to __name__ and just skip all dunder attrs
INTERNAL_CLASS_ATTRS = ['__type__', 'id']
CLASS_ATTRIBUTE_TYPES = (basestring, int, bool, list, float)


class Persistable(object):
    """ The base of all persistable objects.

    Any object stored in the db must inherit from this class.
    Any object having Persistable as it's base are considered persistable.
    """


class StaticClassCollector(object):
    def __init__(self):
        self.reset_state()

    def add_class(self, cls):
        name = get_type_id(cls)
        if name in self.classes:
            raise TypeAlreadyCollected(
                "Type `{}` already defined.".format(name)
            )
        self.classes[name] = cls
        self.descriptors[name] = Descriptor(cls)

    def dump_state(self):
        return self.classes.copy(), self.descriptors.copy()

    def load_state(self, state):
        self.classes, self.descriptors = state
        # make sure we reset all descriptor caches
        for descriptor in self.descriptors.values():
            descriptor._clear_cache()

    def reset_state(self):
        self.load_state(({}, {}))

    def get_descriptors(self):
        return self.descriptors

    def get_classes(self):
        return self.classes


collected_static_classes = StaticClassCollector()


@contextmanager
def collector(type_map=None):
    """ Allows code-defined types to be collected into a custom dict.

    Example:

    with collector() as type_map:
        class Foobar(Entity):
            pass

    type_map == {'Foobar': Foobar}
    """
    if type_map is None:
        type_map = {}

    state = collected_static_classes.dump_state()
    try:

        collected_static_classes.reset_state()
        yield collected_static_classes.get_classes()
    finally:
        collected_static_classes.load_state(state)


class PersistableType(type, Persistable):
    """ Metaclass for static persistable types.

    Collects classes as they are declared so that they can be registered with
    the TypeRegistry later.

    Collection can be controlled using the ``collector`` context manager.
    """
    def __new__(mcs, name, bases, dct):
        if "__type__" in dct:
            raise TypeError("__type__ is a reserved attribute")

        cls = super(PersistableType, mcs).__new__(mcs, name, bases, dct)
        collected_static_classes.add_class(cls)
        return cls


class TypeRegistry(object):
    """ Keeps track of statically and dynamically declared types.
    """
    _builtin_types = (PersistableType,)

    def __init__(self):
        self._dynamic_descriptors = {}
        self._types_in_db = set()

    @property
    def _static_descriptors(self):
        return collected_static_classes.get_descriptors()

    def is_static_type(self, cls):
        class_id = get_type_id(cls)
        descriptor = self._static_descriptors.get(class_id, False)
        return descriptor and descriptor.cls == cls

    def has_code_defined_attribute(self, cls, attr_name):
        """ Determine whether an attribute called ``attr_name`` was defined
        in code on ``cls`` or one of its parent types.
        """
        # find the declaring class in the code-defined hierarchy
        class_id = get_type_id(cls)
        maybe_static = self.get_class_by_id(class_id)
        declaring_cls = get_declaring_class(maybe_static, attr_name,
                                            prefer_subclass=False)
        if not declaring_cls:
            return False  # attribute not found

        # given a declaring class, determine whether that class is dynamic or
        # code-defined
        cls = declaring_cls
        class_id = get_type_id(cls)
        # get_descriptor_by_id will return a dynamically defined class with
        # the given class_id, if there is one. otherwise it falls back to
        # a code-defined class
        maybe_dynamic = self.get_descriptor_by_id(class_id).cls
        # get_class_by_id will return a code-defined class with the given
        # class_id, if there is one. otherwise it falls back to one that
        # has been rehydrated from graph data
        maybe_static = self.get_class_by_id(class_id)

        if maybe_dynamic is maybe_static:
            # type is purely dynamic or purely static
            if not self.is_static_type(maybe_static):
                return False

        is_static_attr = hasattr(maybe_static, attr_name)
        return is_static_attr

    def create_type(self, cls_id, bases, attrs):
        """ Create and register a dynamic type
        """

        with collector():
            cls = PersistableType(cls_id, bases, attrs)

        self.register(cls)

        return cls

    def register(self, cls):
        """ Register a dynamic type
        """
        descriptors = self._dynamic_descriptors

        name = get_type_id(cls)
        if name in descriptors:
            raise TypeAlreadyRegistered(cls)

        descriptors[name] = Descriptor(cls)

    def get_class_by_id(self, cls_id):
        """ Return the class for a given ``cls_id``, preferring statically
        registered classes.

        Returns the statically registered class with the given ``cls_id``, if
        one exists, and otherwise returns any dynamically registered class
        with that id.

        Arguments:
            cls_id: id of the class to return
            registry: type of registered class to prefer

        Returns:
            The class that was registered with ``cls_id``
        """
        try:
            return self._static_descriptors[cls_id].cls
        except KeyError:
            return self.get_descriptor_by_id(cls_id).cls

    def refresh_type(self, cls):
        descriptor = self.get_descriptor(cls)
        descriptor._clear_cache()

    def get_descriptor(self, cls):
        name = get_type_id(cls)
        return self.get_descriptor_by_id(name)

    def get_descriptor_by_id(self, cls_id):
        """ Return the Descriptor for a given ``cls_id``.

        Returns the descriptor for the class registered with the given
        ``cls_id``. If dynamic types have not been loaded yet, return the
        descriptor of the statically registered class.

        Arguments:
            cls_id: id of the class

        Returns:
            The Descriptor for that cls_id

        Raises:
            UnknownType if no type has been registered with the given id.
        """
        if cls_id in self._dynamic_descriptors:
            return self._dynamic_descriptors[cls_id]
        elif cls_id in self._static_descriptors:
            return self._static_descriptors[cls_id]
        else:
            raise UnknownType('Unknown type "{}"'.format(cls_id))

    def get_indexes_for_type(self, cls):
        """
        Find indexes that are used to store instances of the given `cls`.

        Returns:
            A generator of (index_name, key, attr) tuples, where
                `index_name` is the name of the index
                `key` is the index key
                'attr' is the unique attribute object that is the source of
                    values for the index.
        """
        if cls is PersistableType:
            yield(get_index_name(PersistableType), 'id', None)
            return

        descr = self.get_descriptor(cls)
        for name, attr in descr.attributes.items():
            if attr.unique:
                declaring_class = get_declaring_class(descr.cls, name)
                index_name = get_index_name(declaring_class)
                key = name
                yield (index_name, key, attr)

    def get_index_entries(self, obj):
        """
        Find all the index locations that contain `obj`.

        Returns:
            A generator of (index_name, key, value) tuples, each tuple giving
            the indexed location of `obj`.
                `index_name` is the name of the index
                `key` and `value` then represent the location within the index
                    given by `index_name`

                E.g. a response of:
                    ('thing', 'id', 1234),
                    ('another', 'code', 'E108')
                would mean that obj could be found via the following two
                cypher queries:
                    `START n=node:thing("id:1234") RETURN n`
                    `START n=node:another("code:E108") RETURN n`
        """
        if isinstance(obj, PersistableType):
            yield (get_index_name(PersistableType), 'id', obj.__name__)
            return

        obj_type = type(obj)
        for index_name, key, attr in self.get_indexes_for_type(obj_type):
            value = attr.to_primitive(getattr(obj, key), for_db=True)
            if value is not None:
                yield (index_name, key, value)

    def object_to_dict(self, obj, for_db=False):
        """ Converts a persistable object to a dict.

        The generated dict will contain a __type__ key, for which the value
        will be the type_id as given by the descriptor for type(obj).

        If the object is a class, dict will contain an id-key with the value
        being the type_id given by the descriptor for the object.

        For any other object all the attributes as given by the object's
        type descriptpr will be added to the dict and encoded as required.

        Args:
            obj: A persistable  object.

        Returns:
            Dictionary with attributes encoded in basic types
            and type information for deserialization.
            e.g.
            {
                '__type__': 'Entity',
                'attr1' : 1234
            }
        """
        obj_type = type(obj)

        properties = {
            '__type__': get_type_id(obj_type)
        }

        if isinstance(obj, type):
            properties['id'] = get_type_id(obj)
            descr = self.get_descriptor(obj)
            if for_db:
                class_attributes = descr.declared_class_attributes
            else:
                class_attributes = descr.class_attributes

            for name, attr in class_attributes.items():
                # note that we only support native types as class attributes
                properties[name] = getattr(obj, name)

        else:
            descr = self.get_descriptor(obj_type)
            for name, attr in descr.attributes.items():
                try:
                    obj_value = getattr(obj, name)
                    value = attr.to_primitive(obj_value, for_db=for_db)
                except AttributeError:
                    # if we are dealing with an extended type, we may not
                    # have the attribute set on the instance
                    if isinstance(attr, DefaultableAttribute):
                        value = attr.default
                    else:
                        value = None

                if for_db and value is None:
                    continue

                properties[name] = value

        return properties

    def dict_to_object(self, properties):
        """ Converts a dict into a persistable object.

        The properties dict needs at least a __type__ key containing the name
        of any registered class.
        The type key defines the type of the object to return.

        If the registered class for the __type__ is a meta-class,
        i.e. a subclass of <type>, a name key is assumed to be present and
        the registered class idendified by it's value is returned.

        If the registered class for the __type__ is standard class,
        i.e. an instance of <type>, and object of that class will be created
        with attributes as defined by the remaining key-value pairs.

        Args:
            properties: A dict like object.

        Returns:
            A persistable object.
        """

        try:
            type_id = properties['__type__']
        except KeyError:
            raise DeserialisationError(
                'properties "{}" missing __type__ key'.format(properties))

        if type_id == get_type_id(PersistableType):
            # we are looking at a class object
            cls_id = properties['id']
            cls = self.get_class_by_id(cls_id)
            descr = self.get_descriptor_by_id(cls_id)

            for attr_name, value in properties.items():
                if attr_name in INTERNAL_CLASS_ATTRS:
                    continue
                # these are already native (we only support native class attrs)
                setattr(cls, attr_name, value)
            return cls

        # we are looking at an instance object
        cls_id = type_id
        cls = self.get_class_by_id(cls_id)
        obj = cls.__new__(cls)

        descr = self.get_descriptor_by_id(cls_id)

        for attr_name, attr in descr.attributes.items():
            value = properties.get(attr_name)
            value = attr.to_python(value)
            setattr(obj, attr_name, value)

        return obj

    def clone(self):
        """Return a copy of this TypeRegistry that maintains an independent
        dynamic type registry"""
        clone = TypeRegistry()
        clone._dynamic_descriptors = self._dynamic_descriptors.copy()
        clone._types_in_db = self._types_in_db.copy()
        return clone


def get_declaring_class(cls, attr_name, prefer_subclass=True):
    """ Returns the class in the type heirarchy of ``cls`` that defined
        an attribute with name ``attr_name``.

        If ``prefer_subclass`` is false, return the last class in the MRO
        that defined an attribute with the given name.
    """
    declared_attr = None
    declaring_class = None

    # Start at the top of the hierarchy and determine which of the MRO have
    # the attribute. Return the lowest class that defines (or overloads) the
    # attribute, unless prefer_subclass is False.
    for base in reversed(getmro(cls)):
        sentinel = object()
        attr = getattr(base, attr_name, sentinel)
        if attr is not sentinel and declared_attr is not attr:
            declaring_class = base
            declared_attr = attr
            if not prefer_subclass:
                break

    return declaring_class


def get_index_name(cls):
    """ Returns a cypher index name for a class.

    Args:
        cls: The class to generate an index for.

    Returns:
        An index name.
    """
    if issubclass(cls, PersistableType):
        return PersistableType.__name__.lower()
    else:
        return cls.__name__.lower()


def get_type_id(cls):
    """ Returns the type_id for a class.
    """
    if issubclass(cls, PersistableType):
        return PersistableType.__name__
    else:
        return cls.__name__


def is_indexable(cls):
    """ Returns True if the class ``cls`` has any indexable attributes.
    """

    descr = Descriptor(cls)
    for _, attr in descr.attributes.items():
        if attr.unique:
            return True

    return False


def cache_result(func):
    """Cache the result of a function in self._cache
    """
    @wraps(func)
    def decorated(self):
        cache_key = '_{}'.format(func.__name__)

        if cache_key not in self._cache:
            result = func(self)
            self._cache[cache_key] = result
        return self._cache[cache_key]

    return decorated


class Descriptor(object):
    """ Provides information about the types of persistable objects.

    Its main purpose is to provide type names and attribute information of
    persistable types(classes).
    """
    def __init__(self, cls):
        self.cls = cls
        self._cache = {}

    def _clear_cache(self):
        self._cache = {}

    @property
    @cache_result
    def relationships(self):
        from kaiso.attributes.bases import _is_relationship_reference
        relationships = dict(getmembers(
            self.cls, _is_relationship_reference
        ))
        return relationships

    @property
    @cache_result
    def attributes(self):
        attributes = dict(getmembers(self.cls, _is_attribute))

        if issubclass(self.cls, AttributeBase):
            # Because we don't have the attrs on the base Attribute classes
            # declared using Attribute instances, we have to pretend we did,
            # so that they behave like them.
            attributes['name'] = Attribute()
            attributes['unique'] = Attribute()
            attributes['required'] = Attribute()

            if issubclass(self.cls, DefaultableAttribute):
                attributes['default'] = Attribute()

        return attributes

    @property
    @cache_result
    def declared_attributes(self):
        declared = {}
        for name, attr in getmembers(self.cls, _is_attribute):
            if get_declaring_class(self.cls, name) == self.cls:
                declared[name] = attr

        return declared

    @property
    @cache_result
    def class_attributes(self):
        attributes = {}
        for name, value in getmembers(self.cls, _is_class_attribute):
            if name.startswith('__'):
                continue
            attributes[name] = value
        return attributes

    @property
    @cache_result
    def declared_class_attributes(self):
        declared = {}
        for name, value in self.class_attributes.items():
            if get_declaring_class(self.cls, name) == self.cls:
                declared[name] = value

        return declared


class AttributeBase(object):
    __metaclass__ = PersistableType

    name = None

    @staticmethod
    def to_python(value):
        return value

    @staticmethod
    def to_primitive(value, for_db):
        """ Serialize ``value`` to a primitive type suitable for inserting
            into the database or passing to e.g. ``json.dumps``
        """
        return value


def _is_attribute(obj):
    return isinstance(obj, AttributeBase)


def _is_class_attribute(obj):
    return isinstance(obj, CLASS_ATTRIBUTE_TYPES)


class Attribute(AttributeBase):
    unique = None
    required = None

    def __init__(self, unique=False, required=False):
        self.unique = unique
        self.required = required


class DefaultableAttribute(Attribute):
    default = None

    def __init__(self, default=None, **kwargs):
        super(DefaultableAttribute, self).__init__(**kwargs)
        self.default = default

    def __eq__(self, other):
        same_type = type(self) is type(other)
        equal_default = self.default == other.default
        return same_type and equal_default


class AttributedBase(Persistable):
    """ The base class for objects that can have Attributes.

    Sets default values during instance creation and applies kwargs
    passed to __init__.
    """
    __metaclass__ = PersistableType

    def __new__(cls, *args, **kwargs):

        obj = super(AttributedBase, cls).__new__(cls, *args, **kwargs)

        descriptor = Descriptor(cls)

        for name, attr in descriptor.attributes.items():
            setattr(obj, name, attr.default)

        return obj

    def __init__(self, *args, **kwargs):
        super(AttributedBase, self).__init__(*args, **kwargs)

        for key, value in kwargs.items():
            setattr(self, key, value)


class Entity(AttributedBase):
    def __getattribute__(self, name):
        cls = type(self)
        descriptor = Descriptor(cls)
        try:
            rel_reference = descriptor.relationships[name]
        except KeyError:
            return object.__getattribute__(self, name)
        else:
            return rel_reference.get_manager(self)


class Relationship(AttributedBase):
    def __init__(self, start=None, end=None, **kwargs):
        super(Relationship, self).__init__(**kwargs)

        self.start = start
        self.end = end



import pytest

from kaiso.attributes import String, Bool
from kaiso.exceptions import CannotUpdateType, UnsupportedTypeError
from kaiso.persistence import get_type_id
from kaiso.types import Entity, collector


@pytest.fixture
def static_types(manager):
    class Animal(Entity):
        name = String(unique=True)

    class Mammal(Animal):
        pass

    class WaterBound(Animal):
        freshwater = Bool()

    class Cetacean(Mammal):
        pass

    manager.save(Animal)
    manager.save(Mammal)
    manager.save(WaterBound)
    manager.save(Cetacean)

    return {
        'Animal': Animal,
        'Mammal': Mammal,
        'WaterBound': WaterBound,
        'Cetacean': Cetacean,
    }


def test_cannot_reparent_non_type(manager, static_types):
    Animal = static_types['Animal']

    animal = Animal()

    with pytest.raises(UnsupportedTypeError) as e:
        manager.update_type(animal, (Animal,))
    assert e.value.message == "Object is not a PersistableType"


def test_cannot_reparent_code_defined_types(manager, static_types):
    Animal = static_types['Animal']
    Cetacean = static_types['Cetacean']

    manager.save(Cetacean)

    with pytest.raises(CannotUpdateType) as e:
        manager.update_type(Cetacean, (Animal,))
    assert "defined in code" in e.value.message


def test_cannot_reparent_with_incompatible_attributes(manager, static_types):
    Cetacean = static_types['Cetacean']
    WaterBound = static_types['WaterBound']

    Whale = manager.create_type('Whale', (WaterBound,), {})
    manager.save(Whale)

    with pytest.raises(CannotUpdateType) as e:
        manager.update_type(Whale, (Cetacean,))
    assert e.value.message == "Inherited attributes are not identical"

    with collector() as collected:
        class A(Entity):
            foo = String()

        class B(Entity):
            foo = Bool()
    manager.save_collected_classes(collected)

    C = manager.create_type('C', (A,), {})
    manager.save(C)

    with pytest.raises(CannotUpdateType) as e:
        manager.update_type(C, (B,))
    assert e.value.message == "Inherited attributes are not identical"


def test_reparent(manager, static_types):
    """
        (Mammal) <-[:ISA]- (Whale)
    becomes:
        (Mammal) <-[:ISA]- (Cetacean) <-[:ISA]- (Whale)
    """
    Cetacean = static_types['Cetacean']
    Mammal = static_types['Mammal']

    Whale = manager.create_type('Whale', (Mammal,), {})
    manager.save(Whale)

    manager.save(Cetacean)
    manager.update_type(Whale, (Cetacean,))

    UpdatedWhale = manager.type_registry.get_class_by_id(get_type_id(Whale))
    assert UpdatedWhale.__bases__ == (Cetacean,)
    assert issubclass(UpdatedWhale, Cetacean)


def test_reparent_with_subtypes(manager, static_types):
    """
        (Mammal) <-[:ISA]- (Whale) <-[:ISA]- (Orca)
    becomes:
        (Mammal) <-[:ISA]- (Cetacean) <-[:ISA]- (Whale) <-[:ISA]- (Orca)
    """
    Cetacean = static_types['Cetacean']
    Mammal = static_types['Mammal']

    Whale = manager.create_type('Whale', (Mammal,), {})
    Orca = manager.create_type('Orca', (Whale,), {})
    manager.save(Orca)

    manager.save(Cetacean)
    manager.update_type(Whale, (Cetacean,))

    UpdatedWhale = manager.type_registry.get_class_by_id(get_type_id(Whale))
    UpdatedOrca = manager.type_registry.get_class_by_id(get_type_id(Orca))
    assert UpdatedOrca.__bases__ == (UpdatedWhale,)
    assert issubclass(UpdatedOrca, Cetacean)


def test_reparent_with_declared_attributes(manager, static_types):
    """
        (Mammal) <-[:ISA]- (Whale) <-[:ISA]- (Orca) <-[:DECLAREDON]- (name)
    becomes:
        (Mammal) <-[:ISA]- (Cetacean) <-[:ISA]- (Whale) <-[:ISA]- (Orca)
                                                                     ^
                                              (name) -[:DECLAREDON] -'
    """
    Cetacean = static_types['Cetacean']
    Mammal = static_types['Mammal']

    Whale = manager.create_type('Whale', (Mammal,), {})
    Orca = manager.create_type('Orca', (Whale,), {'name': String()})
    manager.save(Orca)

    manager.save(Cetacean)
    manager.update_type(Whale, (Cetacean,))

    UpdatedWhale = manager.type_registry.get_class_by_id(get_type_id(Whale))
    UpdatedOrca = manager.type_registry.get_class_by_id(get_type_id(Orca))
    assert UpdatedOrca.__bases__ == (UpdatedWhale,)
    assert issubclass(UpdatedOrca, Cetacean)

    descriptor = manager.type_registry.get_descriptor(UpdatedOrca)
    assert "name" in descriptor.declared_attributes


def test_reparent_with_matching_attributes(manager):
    """ Reparenting with different but identical attributes is allowed.
    """
    with collector() as collected:
        class C(Entity):
            foo = String()

        class D(Entity):
            foo = String()
    manager.save_collected_classes(collected)

    E = manager.create_type('E', (C,), {})
    manager.save(E)

    manager.update_type(E, (D,))

    UpdatedD = manager.type_registry.get_class_by_id(get_type_id(D))
    UpdatedE = manager.type_registry.get_class_by_id(get_type_id(E))
    assert UpdatedE.__bases__ == (UpdatedD,)
    assert issubclass(UpdatedE, UpdatedD)


def test_reparent_with_instances(manager, static_types):
    """
        (Mammal) <-[:ISA]- (Whale) <-[:ISA]- (Orca) <-[:INSTANCEOF]- (willy)
    becomes:
        (Mammal) <-[:ISA]- (Cetacean) <-[:ISA]- (Orca) <-[:INSTANCEOF]- (willy)
    """
    Cetacean = static_types['Cetacean']
    Mammal = static_types['Mammal']

    Whale = manager.create_type('Whale', (Mammal,), {})
    Orca = manager.create_type('Orca', (Whale,), {})
    willy = Orca(name="willy")

    manager.save(Orca)
    manager.save(willy)

    manager.save(Cetacean)
    manager.update_type(Orca, (Cetacean,))

    UpdatedOrca = manager.type_registry.get_class_by_id(get_type_id(Orca))
    assert UpdatedOrca.__bases__ == (Cetacean,)
    assert issubclass(UpdatedOrca, Cetacean)

    updated_willy = manager.get(Orca, name="willy")
    assert updated_willy.__class__ == UpdatedOrca
    assert isinstance(updated_willy, Cetacean)


def test_reparent_multiple_inheritance(manager):
    """
        (Noun) <-[:ISA]-                      (Verb) <-[:ISA]-
                         \                                     \
                          (Mop)    becomes:                     (Mop)
                         /                                     /
    (Physical) <-[:ISA]-                    (Action) <-[:ISA]-
    """
    Physical = type("Pysical", (Entity,), {})
    Action = type("Action", (Entity,), {})
    Word = type("Word", (Entity,), {})
    Noun = type("Noun", (Word,), {})
    Verb = type("Verb", (Word,), {})

    Mop = manager.create_type('Mop', (Physical, Noun), {})
    manager.save(Mop)

    manager.save(Action)
    manager.save(Verb)
    manager.update_type(Mop, (Action, Verb,))

    UpdatedMop = manager.type_registry.get_class_by_id(get_type_id(Mop))
    assert UpdatedMop.__bases__ == (Action, Verb,)
    assert issubclass(UpdatedMop, Action)
    assert issubclass(UpdatedMop, Verb)
    assert not issubclass(UpdatedMop, (Physical, Noun))


def test_reparent_diamond(manager):
    """
                 -[:ISA]- (Flavouring) <-[:ISA]-
               /                                \
    (Edible) <-                                  - (Beetroot) <-[:ISA]-
               \                               /                       |
                 -[:ISA]- (Colouring) <-[:ISA]-                   (GoldenBeet)
    becomes:
    (Edible) <-[:ISA]- (Vegetable) <-[:ISA]- (Beetroot) <-[:ISA]- (GoldenBeet)
    """
    Edible = type("Edible", (Entity,), {})
    XFlavouring = type("XFlavouring", (Edible,), {})
    XColouring = type("XColouring", (Edible,), {})

    XBeetroot = manager.create_type("XBeetroot", (XFlavouring, XColouring), {})
    GoldenBeet = manager.create_type("GoldenBeet", (XBeetroot,), {})

    Vegetable = type("Vegetable", (Edible,), {})

    manager.save(GoldenBeet)
    manager.save(Vegetable)
    manager.update_type(XBeetroot, (Vegetable,))

    UpdatedBeetroot = manager.type_registry.get_class_by_id(
        get_type_id(XBeetroot))
    UpdatedGoldenBeet = manager.type_registry.get_class_by_id(
        get_type_id(GoldenBeet))
    assert issubclass(UpdatedBeetroot, Vegetable)
    assert issubclass(UpdatedGoldenBeet, UpdatedBeetroot)
    assert not issubclass(UpdatedBeetroot, (XFlavouring, XFlavouring))
    assert not issubclass(UpdatedGoldenBeet, (XFlavouring, XFlavouring))


def test_reparent_missing_type(manager, static_types):
    Cetacean = static_types['Cetacean']
    Mammal = static_types['Mammal']

    Whale = manager.create_type('Whale', (Mammal,), {})

    manager.save(Cetacean)
    with pytest.raises(CannotUpdateType) as e:
        manager.update_type(Whale, (Cetacean,))
    assert e.value.message == "Type or bases not found in the database."


def test_reparent_to_missing_type(manager):
    class Animal(Entity):
        name = String()

    class Mammal(Animal):
        pass

    class Cetacean(Mammal):
        pass

    Whale = manager.create_type('Whale', (Mammal,), {})
    manager.save(Whale)

    with pytest.raises(CannotUpdateType) as e:
        manager.update_type(Whale, (Cetacean,))
    assert e.value.message == "Type or bases not found in the database."



from kaiso.exceptions import MultipleObjectsFound, NoResultFound
from kaiso.references import get_store_for_object

_attribute_types = {}


def wraps_type(cls):
    def wrapper(attr_cls):
        _attribute_types[cls] = attr_cls
        return attr_cls

    return wrapper


def get_attibute_for_type(cls):
    return _attribute_types[cls]


class RelationshipManager(object):
    def __init__(self, obj, relationship_reference):
        self.obj = obj
        self.relationship_reference = relationship_reference

    def _related_objects(self):
        obj = self.obj
        relationship_reference = self.relationship_reference

        store = get_store_for_object(obj)

        related_objects = store.get_related_objects(
            relationship_reference.relationship_class,
            type(relationship_reference),
            obj)

        return related_objects

    def __iter__(self):
        return (rel_obj for rel_obj, _ in self._related_objects())

    @property
    def relationships(self):
        return (rel for _, rel in self._related_objects())

    def first(self):
        return next(iter(self), None)

    def one(self):
        related_objects = iter(self)
        first = next(related_objects, None)
        second = next(related_objects, None)

        if second is not None:
            raise MultipleObjectsFound

        if first is None:
            raise NoResultFound

        return first


def _is_relationship_reference(obj):
    return isinstance(obj, RelationshipReference)


class RelationshipReference(object):
    def __init__(self, relationship_class):
        self.relationship_class = relationship_class

    def get_manager(self, obj):
        return RelationshipManager(obj, self)



from kaiso.attributes.bases import get_attibute_for_type
from kaiso.iter_helpers import unique
from kaiso.relationships import InstanceOf, IsA


def get_changes(old, new):
    """Return a changes dictionary containing the key/values in new that are
       different from old. Any key in old that is not in new will have a None
       value in the resulting dictionary
    """
    changes = {}

    # check for any keys that have changed, put their new value in
    for key, value in new.items():
        if old.get(key) != value:
            changes[key] = value

    for key in old.keys():
        if key not in new:
            raise KeyError('missing key: {}'.format(key))

    return changes


@unique
def get_type_relationships(obj):
    """ Generates a list of the type relationships of an object.
    e.g.
        get_type_relationships(Entity())

        (object, InstanceOf, type),
        (type, IsA, object),
        (type, InstanceOf, type),
        (PersistableType, IsA, type),
        (PersistableType, InstanceOf, type),
        (Entity, IsA, object),
        (Entity, InstanceOf, PersistableType),
        (<Entity object>, InstanceOf, Entity)

    Args:
        obj:    An object to generate the type relationships for.

    Returns:
        A generator, generating tuples
            (object, relatsionship type, related obj)
    """
    obj_type = type(obj)

    if obj_type is not type:
        for item in get_type_relationships(obj_type):
            yield item

    if isinstance(obj, type):
        for base_idx, base in enumerate(obj.__bases__):
            for item in get_type_relationships(base):
                yield item

            yield obj, (IsA, base_idx), base

    yield obj, (InstanceOf, 0), obj_type


def object_to_db_value(obj):
    try:
        attr_cls = get_attibute_for_type(type(obj))
    except KeyError:
        return obj
    else:
        return attr_cls.to_primitive(obj, for_db=True)


def dict_to_db_values_dict(data):
    return dict((k, object_to_db_value(v)) for k, v in data.items())



from kaiso.types import Entity, Relationship, TypeRegistry
from kaiso.queries import get_start_clause
from kaiso.attributes import String


class IndexableThing(Entity):
    indexable_attr = String(unique=True)


class TwoUniquesThing(IndexableThing):
    also_unique = String(unique=True)


class Connects(Relationship):
    indexable_attr = String(unique=True)


type_registry = TypeRegistry()


def test_get_start_clause_for_type():
    clause = get_start_clause(IndexableThing, "foo", type_registry)
    assert clause == 'foo=node:persistabletype(id="IndexableThing")'


def test_get_start_clause_for_instance():
    obj = IndexableThing(indexable_attr="bar")

    clause = get_start_clause(obj, "foo", type_registry)
    assert clause == 'foo=node:indexablething(indexable_attr="bar")'


def test_get_start_clause_mutiple_uniques():
    obj = TwoUniquesThing(
        indexable_attr="bar",
        also_unique="baz"
    )

    clause = get_start_clause(obj, "foo", type_registry)
    assert (clause == 'foo=node:indexablething(indexable_attr="bar")' or
            clause == 'foo=node:indexablething(also_unique="baz")')


def test_get_start_clause_for_relationship_type():
    clause = get_start_clause(Connects, "foo", type_registry)
    assert clause == 'foo=node:persistabletype(id="Connects")'


def test_get_start_clause_for_relationship_instance():
    a = IndexableThing()
    b = IndexableThing()

    obj = Connects(start=a, end=b, indexable_attr="bar")

    clause = get_start_clause(obj, "foo", type_registry)
    assert clause == 'foo=rel:connects(indexable_attr="bar")'



import pytest

from kaiso.exceptions import DeserialisationError
from kaiso.attributes import String, Uuid, Choice
from kaiso.relationships import Relationship, InstanceOf, IsA
from kaiso.serialize import (
    get_type_relationships, get_changes, dict_to_db_values_dict)
from kaiso.types import (
    Persistable, PersistableType, Entity, AttributedBase, Attribute)


def test_classes_to_dict(type_registry):
    class Foo(Entity):
        pass

    dct = type_registry.object_to_dict(Entity)
    assert dct == {'__type__': 'PersistableType', 'id': 'Entity'}

    obj = type_registry.dict_to_object(dct)
    assert obj is Entity

    dct = type_registry.object_to_dict(Foo)
    assert dct == {'__type__': 'PersistableType', 'id': 'Foo'}

    obj = type_registry.dict_to_object(dct)
    assert obj is Foo


def test_objects(type_registry):
    class Foo(Entity):
        pass

    dct = type_registry.object_to_dict(Entity())
    assert dct == {'__type__': 'Entity'}

    obj = type_registry.dict_to_object(dct)
    assert isinstance(obj, Entity)

    dct = type_registry.object_to_dict(Foo())
    assert dct == {'__type__': 'Foo'}

    obj = type_registry.dict_to_object(dct)
    assert isinstance(obj, Foo)


def test_attribute(type_registry):
    """ Verify (de)serialization of Attributes.
    """
    class Bar(Attribute):
        pass

    # Attribute dicts always contain both ``unique`` and ``required`` keys.
    attr = Bar(unique=True)
    dct = type_registry.object_to_dict(attr)
    assert dct == {
        '__type__': 'Bar', 'unique': True, 'required': False, 'name': None}

    # Attribute objects have default values of ``None`` for ``unique``
    # and ``required``.
    obj = type_registry.dict_to_object(
        {'__type__': 'Bar', 'unique': True, 'name': None})
    assert isinstance(obj, Bar)
    assert obj.unique is True
    assert obj.required is None
    assert obj.name is None


def test_choices(type_registry):

    attr = Choice('ham', 'spam', 'eggs')
    dct = type_registry.object_to_dict(attr)

    assert dct == {
        '__type__': 'Choice', 'name': None,
        'unique': False, 'required': False, 'default': None,
        'choices': ['ham', 'spam', 'eggs']}

    obj = type_registry.dict_to_object(dct)
    assert isinstance(obj, Choice)
    assert obj.choices == ('ham', 'spam', 'eggs')


def test_relationship(type_registry):

    dct = type_registry.object_to_dict(Relationship(None, None))
    assert dct == {'__type__': 'Relationship'}

    obj = type_registry.dict_to_object(dct)
    assert isinstance(obj, Relationship)


def test_obj_with_attrs(type_registry):
    class Spam(Entity):
        id = String()
        ham = String(default='eggs')

    spam = Spam(id=None)

    dct = type_registry.object_to_dict(spam, for_db=True)
    assert dct == {'__type__': 'Spam', 'ham': 'eggs'}

    dct = type_registry.object_to_dict(spam, for_db=False)
    assert dct == {'__type__': 'Spam', 'id': None, 'ham': 'eggs'}

    obj = type_registry.dict_to_object(dct)
    assert obj.id == spam.id
    assert obj.ham == spam.ham

    dct.pop('ham')  # removing an attr
    obj = type_registry.dict_to_object(dct)
    assert obj.id == spam.id
    assert obj.ham is None


def test_dynamic_type(type_registry):
    Foobar = type_registry.create_type('Foobar', (Entity,), {})

    dct = type_registry.object_to_dict(Foobar)
    assert dct == {'__type__': 'PersistableType', 'id': 'Foobar'}

    obj = type_registry.dict_to_object(dct)
    assert obj is Foobar


def test_dynamic_typed_object(type_registry):

    attrs = {'id': String()}
    Foobar = type_registry.create_type('Foobar', (Entity,), attrs)

    foo = Foobar(id='spam')
    dct = type_registry.object_to_dict(foo)

    assert dct == {'__type__': 'Foobar', 'id': 'spam'}

    obj = type_registry.dict_to_object(dct)
    assert isinstance(obj, Foobar)
    assert obj.id == foo.id


def test_extend_declared_type_using_dynamic_type(type_registry):

    attrs = {'id': Uuid()}
    DynEntity = type_registry.create_type('Entity', (AttributedBase,), attrs)

    foo = DynEntity()
    dct = type_registry.object_to_dict(foo)

    assert dct == {'__type__': 'Entity', 'id': str(foo.id)}

    obj = type_registry.dict_to_object(dct)
    assert isinstance(obj, Entity)
    assert not isinstance(obj, DynEntity)
    assert obj.id == foo.id


def test_extend_declared_type_using_declared_type(type_registry):
    attrs = {'id': String()}
    type_registry.create_type('Entity', (AttributedBase,), attrs)

    dct = type_registry.object_to_dict(Entity())
    assert dct == {'__type__': 'Entity', 'id': None}


def test_extend_declared_type_with_default_using_declared_type(type_registry):
    attrs = {
        'id': String(default='foobar'),
        'spam': Uuid()
    }
    type_registry.create_type('Entity', (AttributedBase,), attrs)

    dct = type_registry.object_to_dict(Entity())
    assert dct == {'__type__': 'Entity', 'id': 'foobar', 'spam': None}


def test_changes():
    a = {'a': 12, 'b': '34'}
    b = {'a': 1, 'b': '34'}

    changes = get_changes(a, b)
    assert changes == {'a': 1}

    with pytest.raises(KeyError):
        get_changes(a, {})


def test_attribute_types():
    uid = Uuid().default
    dct = dict_to_db_values_dict({'foo': uid, 'bar': 123})

    assert dct == {'foo': str(uid), 'bar': 123}


def test_missing_info(type_registry):
    with pytest.raises(DeserialisationError):
        type_registry.dict_to_object({})


def test_IsA_and_InstanceOf_type_relationships(temporary_static_types):
    """ Testing type hierarchy creation.

    We don't want to test all the types in their relationships,
    but only the set of types which are persisted with IsA and InstanceOf
    relationships.
    """
    class Foo(Entity):
        pass

    result = list(get_type_relationships(Entity))

    assert result == [
        (object, (InstanceOf, 0), type),
        (type, (IsA, 0), object),
        (type, (InstanceOf, 0), type),
        (PersistableType, (IsA, 0), type),
        (Persistable, (IsA, 0), object),
        (Persistable, (InstanceOf, 0), type),
        (PersistableType, (IsA, 1), Persistable),
        (PersistableType, (InstanceOf, 0), type),
        (AttributedBase, (IsA, 0), Persistable),
        (AttributedBase, (InstanceOf, 0), PersistableType),
        (Entity, (IsA, 0), AttributedBase),
        (Entity, (InstanceOf, 0), PersistableType),
    ]

    pers = Entity()
    result = list(get_type_relationships(pers))

    assert result == [
        (object, (InstanceOf, 0), type),
        (type, (IsA, 0), object),
        (type, (InstanceOf, 0), type),
        (PersistableType, (IsA, 0), type),
        (Persistable, (IsA, 0), object),
        (Persistable, (InstanceOf, 0), type),
        (PersistableType, (IsA, 1), Persistable),
        (PersistableType, (InstanceOf, 0), type),
        (AttributedBase, (IsA, 0), Persistable),
        (AttributedBase, (InstanceOf, 0), PersistableType),
        (Entity, (IsA, 0), AttributedBase),
        (Entity, (InstanceOf, 0), PersistableType),
        (pers, (InstanceOf, 0), Entity),
    ]

    foo = Foo()
    result = list(get_type_relationships(foo))

    assert result == [
        (object, (InstanceOf, 0), type),
        (type, (IsA, 0), object),
        (type, (InstanceOf, 0), type),
        (PersistableType, (IsA, 0), type),
        (Persistable, (IsA, 0), object),
        (Persistable, (InstanceOf, 0), type),
        (PersistableType, (IsA, 1), Persistable),
        (PersistableType, (InstanceOf, 0), type),
        (AttributedBase, (IsA, 0), Persistable),
        (AttributedBase, (InstanceOf, 0), PersistableType),
        (Entity, (IsA, 0), AttributedBase),
        (Entity, (InstanceOf, 0), PersistableType),
        (Foo, (IsA, 0), Entity),
        (Foo, (InstanceOf, 0), PersistableType),
        (foo, (InstanceOf, 0), Foo),
    ]



#!/usr/bin/env python
from setuptools import setup, find_packages
from setuptools.command.test import test as TestCommand
from os.path import abspath, dirname, join
import sys


setup_dir = dirname(abspath(__file__))


class PyTest(TestCommand):
    def finalize_options(self):
        TestCommand.finalize_options(self)
        self.test_args = [
            '--cov', 'kaiso',
            '--junitxml=test-results.xml',
            join(setup_dir, 'test'),
        ]
        self.test_suite = True

    def run_tests(self):
        # import here, cause outside the eggs aren't loaded
        import pytest
        errno = pytest.main(self.test_args)
        sys.exit(errno)


def parse_requirements(fn, dependency_links):
    requirements = []
    with open(fn, 'rb') as f:
        for dep in f:
            dep = dep.strip()
            # need to make test_requirements.txt work with
            # setuptools like it would work with `pip -r`
            # URLs will not work, so we transform them to
            # dependency_links and requirements
            if dep.startswith('git+'):
                dependency_links.append(dep)
                _, dep = dep.rsplit('#egg=', 1)
                dep = dep.replace('-', '==', 1)
            requirements.append(dep)

    return requirements, dependency_links

requirements, dependency_links = parse_requirements(
    join(setup_dir, 'requirements.txt'), [])

test_requirements, dependency_links = parse_requirements(
    join(setup_dir, 'test_requirements.txt'),
    dependency_links)

setup(
    name='kaiso',
    version='0.16.2',
    description='A queryable object persistence and relationship framework'
                ' based on the Neo4j graph database.',
    author='onefinestay',
    author_email='engineering@onefinestay.com',
    url='http://github.com/onefinestay/kaiso',
    packages=find_packages(exclude=['test', 'test.*']),
    install_requires=requirements,
    tests_require=test_requirements,
    dependency_links=dependency_links,
    zip_safe=True,
    cmdclass={'test': PyTest},
    license='Apache License, Version 2.0',
    classifiers=[
        "Programming Language :: Python",
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: POSIX",
        "Programming Language :: Python :: 2.7",
        "Topic :: Internet",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Intended Audience :: Developers"
    ]
)



import pytest

from kaiso.exceptions import UnknownType
from kaiso.migration_helpers import validate_base_change
from kaiso.types import collector, Entity


def test_basic(manager):
    with collector() as collected:
        class A(Entity):
            pass

        class A2(A):
            pass

        class B(Entity):
            pass

        class B2(B):
            pass

        class AB(A, B):
            pass

    manager.save_collected_classes(collected)
    validate_base_change(manager, 'AB', ('A2', 'B2'))


def test_become_your_own_ancestor(manager):
    with collector() as collected:
        class A(Entity):
            pass

        class A2(A):
            pass

        class A3(A2):
            pass

    manager.save_collected_classes(collected)

    # become your own parent
    with pytest.raises(ValueError) as ex:
        validate_base_change(manager, 'A', ('A3',))
    assert "inheritance cycle" in str(ex)


def test_duplicate_base_class(manager):
    with collector() as collected:
        class A(Entity):
            pass

        class B(Entity):
            pass

        class C(A, B):
            pass

    manager.save_collected_classes(collected)

    with pytest.raises(ValueError) as ex:
        validate_base_change(manager, 'C', ('A', 'B', 'A'))
    assert "duplicate base class" in str(ex)


def test_move_down_the_hieararchy(manager):

    with collector() as collected:
        class A(Entity):
            pass

        class A2(A):
            pass

        class A3(A2):
            pass

        class A4(A3):
            pass

        class B(A):
            pass

        class C(B):
            pass

    manager.save_collected_classes(collected)

    validate_base_change(manager, 'B', ('A4',))


def test_bad_mro(manager):
    """
    from http://www.python.org/download/releases/2.3/mro/

     -----------
    |           |
    |    O      |
    |  /   \    |
     - X    Y  /
       |  / | /
       | /  |/
       A    B
       \   /
         ?
    """

    with collector() as collected:
        class X(Entity):
            pass

        class Y(Entity):
            pass

        class A(X, Y):
            pass

        class B(Y):  # to become B(Y, X)
            pass

        class AB(A, B):
            pass

    manager.save_collected_classes(collected)

    with pytest.raises(ValueError) as ex:
        validate_base_change(manager, 'B', ('Y', 'X'))
    assert "Cannot create a consistent method resolution" in str(ex)


def test_unknown_types(manager):
    with collector() as collected:
        class A(Entity):
            pass

    manager.save_collected_classes(collected)

    with pytest.raises(UnknownType):
        validate_base_change(manager, 'A', ['Invalid'])

    with pytest.raises(UnknownType):
        validate_base_change(manager, 'Invalid', ['A'])



from kaiso.types import collected_static_classes
from kaiso.types import PersistableType, Entity
from kaiso.attributes import String


def test_class_collection():

    class CollectFoo():
        __metaclass__ = PersistableType

    class CollectBar(Entity):
        baz = String()

    assert "CollectFoo" in collected_static_classes.get_classes()
    assert "CollectBar" in collected_static_classes.get_classes()
    assert "Entity" in collected_static_classes.get_classes()
    assert "String" in collected_static_classes.get_classes()



import pytest

from kaiso.types import Entity
from kaiso.test_helpers import TemporaryStaticTypes


@pytest.fixture
def bar(temporary_static_types):
    class Bar(Entity):
        pass


def test_context_manager_a():
    with TemporaryStaticTypes():
        class Bar(Entity):
            pass


def test_context_manager_b():
    with TemporaryStaticTypes():
        class Bar(Entity):
            pass


def test_fixture_a(temporary_static_types):
    class Bar(Entity):
        pass


def test_fixture_b(temporary_static_types):
    class Bar(Entity):
        pass


def test_fixture_with_types_a(bar):
    pass


def test_fixture_with_types_b(bar):
    pass



""" Provides a connection factory for connecting to existing or
temporary neo4j instances.

For temporary instances it requires the neo4j command
to be available on the path.
"""
import os
import time
import atexit
import logging
import requests
import tempfile
import urlparse
import subprocess

from collections import namedtuple

from py2neo import neo4j

from kaiso.exceptions import ConnectionError

logger = logging.getLogger(__name__)

temp_neo4j = namedtuple('temp_neo4j', ['http_url', 'process'])
_temporary_databases = {}

TIMEOUT = 30  # seconds
DEFAULT_TEMP_PORT = 7475


def get_neo4j_info():
    """ Gets runtime information from the neo4j command.

    Returns:
        A dict.
    """
    default_cmd = os.pathsep.join([
        'neo4j',
        '/var/lib/neo4j/bin/neo4j',  # >= 1.9.4
        '/etc/init.d/neo4j-service',
    ])
    neo4j_cmds = os.environ.get('NEO4J_CMD', default_cmd).split(os.pathsep)
    output = None
    for cmd in neo4j_cmds:
        try:
            output = subprocess.check_output([cmd, 'info'])
        except (OSError, subprocess.CalledProcessError):
            pass

    if not output:
        raise ConnectionError('Cannot determine neo4j info. Is the NEO4J_CMD '
                              'environment varaible set correctly?')

    keys = ['NEO4J_HOME', 'NEO4J_INSTANCE', 'JAVA_OPTS', 'CLASSPATH']

    result = {}

    for lne in output.splitlines():
        for key in keys:
            sep = key + ':'
            if lne.startswith(sep):
                _, value = lne.split(sep)
                value = value.strip()
                result[key] = value

    return result


def write_config(config_filepath, port, temp_data_dir, bind_iface):
    """ Write neo4j server properties into the file at ``config_filepath``
    """
    config_options = {
        'org.neo4j.server.webserver.port': port,
        'org.neo4j.server.database.location': temp_data_dir,
        'org.neo4j.server.webserver.address': bind_iface
    }

    with open(config_filepath, 'w') as props_file:
        for key, value in config_options.iteritems():
            props_file.write("{}={}\n".format(key, value))


def temp_neo4j_instance(uri):
    """ Start or return an existing instance of the neo4j graph database server
    using the given URI.
    """
    # split the uri
    split_uri = urlparse.urlparse(uri)
    port = split_uri.port or DEFAULT_TEMP_PORT
    bind_iface = split_uri.hostname or "localhost"

    # return http uri if this temporary database already exists
    if uri in _temporary_databases:
        return _temporary_databases[uri].http_url

    neo4j_info = get_neo4j_info()

    temp_dir = split_uri.path or tempfile.mkdtemp()
    temp_data_dir = os.path.join(temp_dir, 'data')
    if not os.path.exists(temp_data_dir):
        os.makedirs(temp_data_dir)

    # otherwise, start a new neo4j process.
    # define the subprocess command and the required classpath
    cmd = ['java', '-cp']
    cmd.append(neo4j_info['CLASSPATH'])

    # neo requires a physical config file which we temporarily create to
    # specify config overrides
    props_filepath = os.path.join(temp_dir, 'server.properties')
    write_config(props_filepath, port, temp_data_dir, bind_iface)

    # required startup args
    args = [
        '-server', '-XX:+DisableExplicitGC'
    ]

    # additional startup options for the command line
    startup_options = {
        'neo4j.home': neo4j_info['NEO4J_HOME'],
        'neo4j.instance': neo4j_info['NEO4J_INSTANCE'],
        'org.neo4j.server.properties': props_filepath,
        'file.encoding': 'UTF-8',
        'java.awt.headless': 'true',
    }
    for key, value in startup_options.iteritems():
        args.append("-D{}={}".format(key, value))

    # finally add the class to run
    args.append('org.neo4j.server.Bootstrapper')
    cmd.extend(args)

    # start the subprocess
    neo4j_process = subprocess.Popen(cmd, stdout=subprocess.PIPE,
                                     stderr=subprocess.STDOUT)

    # terminate subprocess at exit
    def terminate():  # pragma: no cover
        try:
            neo4j_process.terminate()
        except OSError:
            pass  # already terminated
    atexit.register(terminate)

    # the startup process is async so we monitor the http interface to know
    # when to allow the test runner to continue
    timeout_time = time.time() + TIMEOUT
    hostname = "localhost" if bind_iface == "0.0.0.0" else bind_iface
    http_url = "http://{}:{}/db/data/".format(hostname, port)

    while time.time() < timeout_time:
        try:
            req = requests.get(http_url)
            if "neo4j_version" in req.text:
                logger.debug('neo4j server started on {}'.format(http_url))
                _temporary_databases[uri] = temp_neo4j(http_url, neo4j_process)
                return http_url  # return REST API url
        except requests.ConnectionError:
            time.sleep(0.2)

    logger.critical(
        'Unable to start Neo4j: timeout after %s '
        'seconds. See logs in %s.', TIMEOUT, temp_data_dir)

    raise ConnectionError(http_url)


def get_connection(uri):
    if not uri:
        raise ConnectionError("You must provide a connection URI")
    if uri.startswith('temp://'):
        uri = temp_neo4j_instance(uri)
    graph_db = neo4j.GraphDatabaseService(uri)
    return graph_db
